import StringIO

from . import rtp


clock_rate = 48000


class OpusRTPPayload(rtp.RTPAudioPayloadMixin, rtp.RTPPayload):

    def __init__(self, *args, **kwargs):
        if args:
            if kwargs:
                raise TypeError('Mixed positional and keyword arguments')
            elif len(args) > 1:
                raise TypeError('Expected single \'buf\' arg')
            self.unpack(args[0])
        elif kwargs:
            if len(kwargs) == 1 and 'buf' in kwargs:
                self.unpack(kwargs['buf'])
            else:
                self.data = ''
                for k, v in kwargs.iteritems():
                    if k not in ('data',):
                        raise TypeError('Unexpected keyword argument \'{0}\''.format(k))
                    setattr(self, k, v)

    @property
    def nb_frames(self):
        count = ord(self.data[0]) & 0x3;
        if count == 0:
            return 1
        elif count != 3:
            return 2
        elif len(self.data) < 2:
            raise ValueError('Invalid packet')
        return ord(self.data[1]) & 0x3F;

    @property
    def nb_samples_per_frame(self):
        toc = ord(self.data[0])
        size = 0
        if toc & 0x80:
            size = ((toc >> 3) & 0x3);
            size = (clock_rate << size) / 400;
        elif toc & 0x60 == 0x60:
            size = clock_rate / 50 if toc & 0x08 else clock_rate / 100;
        else:
            size = ((toc >> 3) & 0x3)
            if size == 3:
                size = clock_rate * 60 / 1000
            else:
                size = (clock_rate << size) / 100
        return size

    # RTPAudioMixin

    @property
    def nb_samples(self):
        """
        https://github.com/xiph/opus/blob/5dca296833ce4941dceadf956ff0fb6fe59fe4e8/src/opus_decoder.c#L960
        """
        samples = self.nb_frames * self.nb_samples_per_frame
        # NOTE: can't have more than 120 ms
        if samples * 25 > clock_rate * 3:
            raise ValueError('Invalid packet')
        return samples

    @property
    def nb_channels(self):
        """
        https://github.com/xiph/opus/blob/5dca296833ce4941dceadf956ff0fb6fe59fe4e8/src/opus_decoder.c#L939
        """
        return 2 if (ord(self.data[0]) & 0x4) else 1

    # RTPPayload

    def pack(self, fo=None):
        if fo is None:
            fo, value = StringIO.StringIO(), True
        else:
            value = False
        fo.write(self.data)
        if value:
            return fo.getvalue()

    def unpack(self, buf):
        self.data = buf


class OpusRTPPacket(rtp.RTPTimeMixin, rtp.RTPPacket):
    """
    https://tools.ietf.org/html/draft-ietf-payload-rtp-opus-11
    """

    # RTPPacket

    type = rtp.RTPPacket.AUDIO_TYPE

    payload_type = OpusRTPPayload

    # RTPTimeMixin

    clock_rate = clock_rate



import inspect

import pytest

import marm


@pytest.mark.parametrize(
    ('stored,pt,ssrc,packet_type,min_window,frame_rate'), [
        ('sonic-v.mjr',
         100,
         1653789901,
         marm.vp8.VP8RTPPacket,
         10,
         31),
        ('streets-of-rage.pcap',
         100,
         3830765780,
         marm.vp8.VP8RTPPacket,
         10,
         21),
        ('streets-of-rage.pcap',
         100,
         3830765780,
         marm.vp8.VP8RTPPacket,
         10 ** 10,
         ValueError),
    ])
def test_rtp_estimate_video_frame_rate(
        fixtures,
        stored,
        pt,
        ssrc,
        packet_type,
        min_window,
        frame_rate):
    s_path = fixtures.join(stored)
    with s_path.open() as fo:
        pkts = marm.rtp.RTPPacketReader.open(
            fo,
            packet_filter=lambda pkt: (
                pkt.header.type == pt and pkt.header.ssrc == ssrc
            ),
            packet_type=packet_type,
        )
        if inspect.isclass(frame_rate) and issubclass(frame_rate, Exception):
            with pytest.raises(frame_rate):
                marm.rtp.estimate_video_frame_rate(pkts, min_window=min_window)
        else:
            assert int(
                marm.rtp.estimate_video_frame_rate(pkts, min_window=min_window)
            ) == frame_rate


@pytest.mark.parametrize(
    ('stored,pt,ssrc,packet_type,width,height'), [
        ('sonic-v.mjr', 100, 1653789901, marm.vp8.VP8RTPPacket, 320, 240),
        ('streets-of-rage.pcap', 100, 3830765780, marm.vp8.VP8RTPPacket, 960,
         720),
    ])
def test_rtp_video_dimensions(
        fixtures,
        stored,
        pt,
        ssrc,
        packet_type,
        width,
        height):
    s_path = fixtures.join(stored)
    with s_path.open() as fo:
        pkts = marm.rtp.RTPPacketReader.open(
            fo,
            packet_filter=lambda pkt: (
                pkt.header.type == pt and pkt.header.ssrc == ssrc
            ),
            packet_type=packet_type,
        )
        assert marm.rtp.probe_video_dimensions(pkts) == (width, height)


@pytest.mark.parametrize(
    ('v_store,v_count,v_p_kf, v_n_kf,a_store,a_count,dur,parts'), [
        ('sonic-v.mjr', 9654, [
            ((False, (0, 0), None, (None, None))),
            ((True, (0, 164), -5.899999999999636, (0, 204))),
            ((True, (1, 462), -0.8659999999999854, (1, 456))),
            ((True, (1, 462), -10.865999999999985, (1, 456))),
            ((True, (3, 250), -5.899999999999636, (3, 204))),
            ((True, (4, 621), -0.7999999999992724, (4, 460))),
            ((True, (4, 621), -10.799999999999272, (4, 460))),
            ((True, (6, 312), -5.798999999999978, (6, 210))),
            ((True, (7, 931), -0.7999999999992724, (7, 460))),
            ((True, (7, 931), -10.799999999999272, (7, 460))),
            ((True, (9, 574), -0.9659999999994398, (9, 451))),
            ((True, (10, 394), -6.8659999999999854, (10, 156))),
         ], [
            ((True, (0, 4), 0.052999999999883585, (0, 3))),
            ((True, (1, 462), 9.134000000000015, (1, 457))),
            ((True, (3, 250), 14.100000000000364, (3, 205))),
            ((True, (3, 250), 4.100000000000364, (3, 205))),
            ((True, (4, 621), 9.200000000000728, (4, 460))),
            ((True, (6, 312), 14.201000000000022, (6, 211))),
            ((True, (6, 312), 4.201000000000022, (6, 211))),
            ((True, (7, 931), 9.200000000000728, (7, 460))),
            ((True, (9, 403), 14.234000000000378, (9, 212))),
            ((True, (9, 403), 4.234000000000378, (9, 212))),
            ((True, (10, 2), 0.06700000000000728, (10, 4))),
            ((True, (11, 1069), 8.167000000000371, (11, 409))),
         ],
         'sonic-a.mjr', 5996,
         10.0, 12),
    ])
def test_rtp_cursor(
        tmpdir, fixtures,
        v_store, v_count, v_p_kf, v_n_kf,
        a_store, a_count,
        dur, parts):
    # split v
    v_path = fixtures.join(v_store)
    v_parts = 0
    v_pkts = marm.mjr.MJRRTPPacketReader(
        str(v_path),
        packet_type=marm.vp8.VP8RTPPacket
    )
    for i, part in enumerate(marm.rtp.split_packets(v_pkts, duration=dur)):
        t_path = tmpdir.join('v-{0}.mjr'.format(i + 1))
        with t_path.open('wb') as t_fo:
            marm.mjr.write_header(t_fo, 'video')
            for pkt in part:
                marm.mjr.write_packet(t_fo, pkt)
        v_parts += 1
    assert v_parts == parts

    # split a
    a_path = fixtures.join(a_store)
    a_parts = 0
    a_pkts = marm.mjr.MJRRTPPacketReader(
        str(a_path),
        packet_type=marm.opus.OpusRTPPacket
    )
    for i, part in enumerate(marm.rtp.split_packets(a_pkts, duration=dur)):
        t_path = tmpdir.join('a-{0}.mjr'.format(i + 1))
        with t_path.open('wb') as t_fo:
            marm.mjr.write_header(t_fo, 'audio')
            for pkt in part:
                marm.mjr.write_packet(t_fo, pkt)
        a_parts += 1
    assert a_parts == parts

    # cursors
    v_cur = marm.rtp.RTPCursor([
            str(tmpdir.join('v-{0}.mjr'.format(i + 1)))
            for i in range(v_parts)
        ],
        marm.mjr.MJRRTPPacketReader,
        packet_type=marm.vp8.VP8RTPPacket,
    )
    a_cur = marm.rtp.RTPCursor([
            str(tmpdir.join('a-{0}.mjr'.format(i + 1)))
            for i in range(a_parts)
        ],
        marm.mjr.MJRRTPPacketReader,
        packet_type=marm.opus.OpusRTPPacket,
    )

    # count packets
    assert sum(1 for _ in v_cur) == v_count
    assert sum(1 for _ in v_cur) == 1  # exhausted
    assert sum(1 for _ in a_cur) == a_count
    assert sum(1 for _ in a_cur) == 1  # exhausted

    # prev key-frames
    for i in range(v_parts):
        hit, v_tell, v_delta, a_tell = v_p_kf[i]
        v_cur.seek((i, 0))
        s_pkt = v_cur.current()
        pkt = v_cur.prev_key_frame()
        assert v_cur.tell() == v_tell
        if hit:
            assert pkt is not None and pkt.data.is_key_frame
            delta = pkt.secs - s_pkt.secs
            assert v_delta == delta
            a_cur.seek((i, 0))
            a_cur.rewind(-v_delta)
            assert a_cur.tell() == a_tell
        else:
            assert pkt is None

    # next key-frames
    for i in range(v_parts):
        (hit, v_tell, v_delta, a_tell) = v_n_kf[i]
        v_cur.seek((i, 0))
        s_pkt = v_cur.current()
        pkt = v_cur.next_key_frame()
        assert v_cur.tell() == v_tell
        if hit:
            assert pkt is not None and pkt.data.is_key_frame
            delta = pkt.secs - s_pkt.secs
            assert v_delta == delta
            a_cur.seek((i, 0))
            a_cur.fastforward(v_delta)
            assert a_cur.tell() == a_tell


@pytest.mark.parametrize(
    ('store,samples'), [
        ('sonic-a.mjr', 5756160),
        ('streets-of-rage.pcap', 706920),
    ],
)
def test_rtp_audio_samples(fixtures, tmpdir, store, samples):
    p = fixtures.join(store)
    pkts = marm.rtp.RTPPacketReader.open(
        p.strpath, packet_type=marm.opus.OpusRTPPacket,
    )
    assert sum(pkt.data.nb_samples for pkt in pkts) == samples



"""
Command line interface to marm. See:

.. code:: bash

    $ marm --help

for usage.
"""
import argparse
import collections
import logging
import os
import re

from . import __version__, rtp, vp8, opus, mjr, frame, Frames, VideoFrame, VideoFrames


logger = logging.getLogger(__name__)


packet_types = {
    'vp8': vp8.VP8RTPPacket,
    'opus': opus.OpusRTPPacket,
}

encoders = {
    vp8.VP8RTPPacket: 'libvpx',
    opus.OpusRTPPacket: 'libopus',
}

log_levels = {
    'd': logging.DEBUG,
    'i': logging.INFO,
    'w': logging.WARN,
    'e': logging.ERROR,
}


class StreamDescription(collections.namedtuple(
        'StreamDescription', ['packet_types', 'order']
    )):

    INTERLEAVE = 'interleave'
    CONCAT = 'concat'
    
    def bucket(self, *archives):
        if len(archives) == 1 and isinstance(archives[0], list):
            archives = archives[0]
        else:
            archives = list(archives)
        if len(archives) % len(self.packet_types) != 0:
            raise ValueError(
                'Archive count must be a factor of {0}'
                .format(len(self.packet_types))
            )
        if self.order == self.INTERLEAVE:
            s = len(self.packet_types)
            buckets = [
                (self.packet_types[i], archives[i::s])
                for i in range(s)
            ]
        else:  # self.order == self.CONCAT
            n = len(archives) / len(self.packet_types)
            buckets = [
                (self.packet_types[i / n], archives[i:i + n])
                for i in range(0, len(archives), n)
            ]
        return buckets


class ParamsAction(argparse.Action):
    
    params = None
    
    def parse(self, value):
        profile = {}
        for k, v in re.findall(r'([^=,\s]*)\s*=\s*([^,]*)', value):
            if k not in self.params:
                raise argparse.ArgumentError(
                    self, '{0}= not a supported param'.format(k),
                )
            try:
                pv = self.params[k](v)
            except (TypeError, ValueError), ex:
                raise argparse.ArgumentError(
                    self, '{0}= value "{1}" invalid ({2})'.format(k, v, ex),
                )
            profile[k] = pv
        return profile


    # argparse.Action

    def __call__(self, parser, namespace, values, option_string=None):
        if isinstance(values, basestring):
            setattr(namespace, self.dest, self.parse(values))
        else:
            setattr(namespace, self.dest, map(self.parse, values))


class PacketFilterAction(ParamsAction):

    # ParamsAction

    params = {
        'pt': int,
        'ssrc': int,
    }


class VideoProfileAction(ParamsAction):
    
    # ParamsAction

    params = {
        'encoder_name': str,
        'pix_fmt': int,
        'width': int,
        'height': int,
        'frame_rate': float,
        'bit_rate': int,
    }


class AudioProfileAction(ParamsAction):

    # ParamsAction

    params = {
        'encoder_name': str,
        'sample_rate': int,
        'bit_rate': int,
    }


class StreamDescriptionAction(argparse.Action):

    def parse(self, value):
        if ',' in value:
            types = value.split(',')
            order = StreamDescription.CONCAT
        elif '!' in value:
            types = value.split('!')
            order = StreamDescription.INTERLEAVE
        else:
            types = [value]
            order = StreamDescription.CONCAT
        for t in types:
            if t not in packet_types:
                raise argparse.ArgumentError(
                    self, '"{0}" not a supported packet type'.format(t),
                )
        return StreamDescription(
            packet_types=[packet_types[t] for t in types],
            order=order
        )
    
    # argparse.Action
    
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, map(self.parse, values))


def split_parser(cmd_parsers, parents):
    """
    Split command parser.
    """
    split_parser = cmd_parsers.add_parser(
        'split',
        help='splits archived media packets into multiple files',
        parents=parents,
    )
    split_parser.add_argument(
        '-d', '--duration', '--dur',
        type=float,
        default=None,
        help='limit duration of each part to this',
        metavar='DURATION',
    )
    split_parser.add_argument(
        '-c', '--count',
        type=int,
        default=None,
        help='limit number of packets in each part to this',
        metavar='COUNT',
    )
    split_parser.add_argument(
        '-f', '--force',
        action='store_true',
        default=False,
        help='overwrite existing splits',
    )
    split_parser.add_argument(
        '--filter',
        action=PacketFilterAction,
        help='packet filter',
    )
    split_parser.add_argument(
        'packet_type',
        choices=packet_types.keys(),
        help='type of media packet in archive',
        metavar='PACKET-TYPE',
    )
    split_parser.add_argument(
        'in_path',
        nargs=1,
        help='path to input archive'
    )
    split_parser.add_argument(
        'out_format',
        nargs='?',
        help='format for path to split archive (e.g. /tmp/out-{part:02}).'
    )
    split_parser.set_defaults(cmd=split_cmd)


def split_cmd(args):
    """
    Split command.
    """
    in_path = args.in_path[0]
    out_format = args.out_format
    if out_format is None or os.path.isdir(out_format):
        p = os.path.abspath(in_path)
        _, ext = os.path.splitext(p)
        name = os.path.basename(in_path)[:-len(ext)]
        if ext == '.pcap':
            ext = '.mjr'
        d = out_format if out_format else os.path.dirname(p)
        out_format = os.path.join(d, '{0}-{1}{2}'.format(name, '{part:02}', ext))
        logger.info('generated format "%s"', out_format)
    _, ext = os.path.splitext(out_format)
    if ext:
        ext = ext[1:]
    if ext != 'mjr':
        raise ValueError('Only {0} output archives supported.'.format('mjr'))

    logger.info(
        'splitting "%s" w/ duration=%s, count=%s',
        in_path, args.duration, args.count,
    )
    if args.filter:
        pt = args.filter.get('pt')
        ssrc = args.filter.get('ssrc')
        packet_filter = lambda pkt: (
            (pt is None or pkt.header.type == pt) and
            (ssrc is None or pkt.header.ssrc == ssrc)
        )
    else:
        packet_filter = None
    pkts = rtp.RTPPacketReader.open(
        in_path,
        packet_type=packet_types[args.packet_type],
        packet_filter=packet_filter,
    )
    splits = rtp.split_packets(pkts, duration=args.duration, count=args.count)
    split_count = 0
    pkt_total = 0
    for i, split in enumerate(splits):
        out_path = out_format.format(part=i + 1)
        if not args.force and os.path.exists(out_path):
            pkt_count = sum(1 for _ in split)
            logger.warn(
                'not overwriting existing split %s @ "%s", skipping %s packets',
                i + 1, out_path, pkt_count,
            )
        else:
            logger.info('writing split %s to "%s"', i + 1, out_path)
            pkt_count = 0
            with open(out_path, 'wb') as fo:
                mjr.write_header(fo, pkts.packet_type.type)
                for pkt in split:
                    mjr.write_packet(fo, pkt)
                    pkt_count += 1
            logger.info('wrote %s packets to split "%s"', pkt_count, out_path)
            pkt_total += pkt_count
            split_count += 1
    logger.info(
        'wrote %s packets to %s splits w/ format "%s"',
        pkt_total, split_count, out_format,
    )


def mux_parser(cmd_parsers, parents):
    """
    Mux command parser.
    """
    mux_parser = cmd_parsers.add_parser(
        'mux',
        help='multiplexes archived media packets to a container file',
        parents=parents,
    )
    mux_parser.add_argument(
        '-o', '--offset',
        type=int,
        default=0,
        help='offset to stream',
        metavar='OFFSET',
    )
    mux_parser.add_argument(
        '-v', '--video-profile',
        action=VideoProfileAction,
        default=[],
        help='video profile to use when muxing',
    )
    mux_parser.add_argument(
        '-a', '--audio-profile',
        action=AudioProfileAction,
        default=[],
        help='audio profile to use when muxing',
    )
    mux_parser.add_argument(
        '-f', '--force',
        action='store_true',
        default=False,
        help='overwrite existing container',
    )
    mux_parser.add_argument(
        '-d', '--dry',
        action='store_true',
        default=False,
        help='abort just before muxing',
    )
    mux_parser.add_argument(
        '--video-frame-rate-window',
        type=int,
        default=10,
        help='number of video frames over which estimate frame rate',
    )
    mux_parser.add_argument(
        'container',
        nargs=1,
        help='mux input archives to this container file',
    )
    mux_parser.add_argument(
        'description',
        action=StreamDescriptionAction,
        nargs=1,
        help=(
            'packet-type(s) ({0}) separated by , (interleaved) or | (grouped) '
            'describing input archives to follow'
            .format(', '.join(packet_types.keys()))
        ),
        metavar='DESCRIPTION',
    )
    mux_parser.add_argument(
        'archives',
        nargs='*',
        help=(
            'input archive w/ packet type and bucketing controlled by '
            'DESCRIPTION'
        ),
    )
    mux_parser.set_defaults(cmd=mux_cmd)


def mux_cmd(args):
    """
    Mux command.
    """
    description = args.description[0]
    logger.info(
        'bucketing archives to cursors using description %s', description
    )
    buckets = description.bucket(args.archives)
    for i, (_, parts) in enumerate(buckets):
        logger.info(
            'stream %s (%s) - \n%s',
            i,
            description.packet_types    [i],
            '\n'.join('  {0}'.format(part) for part in parts),
        )
    curs = [
        rtp.RTPCursor(
            archives,
            rtp.RTPPacketReader.open,
            packet_type=packet_type,
        )
        for packet_type, archives in buckets
    ]

    if args.offset:
        logger.info('offsetting cursors to %s', args.offset)
        for cur in curs:
            cur.seek((args.offset, 0))

    logger.info('resolving video')
    v_curs = [
        (i, packet_type, curs[i])
        for i, (packet_type, _) in enumerate(buckets)
        if packet_type.type == rtp.RTPPacket.VIDEO_TYPE
    ]
    if len(v_curs) > 1:
        raise NotImplementedError('Only one video stream for now')
    if v_curs:
        i, packet_type, v_cur = v_curs[0]
        v_prof = (
            args.video_profile[i].copy()
            if i < len(args.video_profile)
            else {}
        )
        if 'encoder_name' not in v_prof:
            v_prof['encoder_name'] = encoders[packet_type]
        if 'pix_fmt' not in v_prof:
            v_prof['pix_fmt'] = VideoFrame.PIX_FMT_YUV420P
        if 'width' not in v_prof:
            with v_cur.restoring():
                width, _ = rtp.probe_video_dimensions(v_cur)
            v_prof['width'] = width
        if 'height' not in v_prof:
            with v_cur.restoring():
                _, height = rtp.probe_video_dimensions(v_cur)
            v_prof['height'] = height
        if 'frame_rate' not in v_prof:
            with v_cur.restoring():
                frame_rate = rtp.estimate_video_frame_rate(
                    v_cur, window=args.video_frame_rate_window
                )
            v_prof['frame_rate'] = frame_rate
        if 'bit_rate' not in v_prof:
            v_prof['bit_rate'] = 1000000
        v_prof['time_base'] = (1, 1000)
        logger.info(
            'using video profile -\n%s',
            '\n'.join('  {0}={1}'.format(k, v) for k, v in v_prof.items())
        )
        v_frames = VideoFrames(v_cur)
    else:
        logger.info('no video')
        v_frames = None
        v_prof = None

    logger.info('resolving audio')
    a_curs = [
        (i, packet_type, curs[i])
        for i, (packet_type, _) in enumerate(buckets)
        if packet_type.type == rtp.RTPPacket.AUDIO_TYPE
    ]
    if len(a_curs) > 1:
        raise NotImplementedError('Only one audio stream for now')
    if a_curs:
        i, packet_type, a_cur = a_curs[0]
        a_prof = (
            args.audio_profile[i].copy()
            if i < len(args.audio_profile)
            else {}
        )
        if 'encoder_name' not in a_prof:
            a_prof['encoder_name'] = encoders[packet_type]
        if 'sample_rate' not in a_prof:
            a_prof['sample_rate'] = 48000
        if 'bit_rate' not in a_prof:
            a_prof['bit_rate'] = 96000
        if 'channel_layout' not in a_prof:
            with v_cur.restoring():
                a_prof['channel_layout'] = rtp.probe_audio_channel_layout(a_cur)
        a_prof['time_base'] = (1, 1000)
        logger.info(
            'using audio profile -\n%s',
            '\n'.join('  {0}={1}'.format(k, v) for k, v in a_prof.items())
        )
        a_frames = Frames(a_cur)
    else:
        logger.info('no audio')
        a_frames = None
        a_prof = None

    container = args.container[0]
    logger.info('muxing frames to "%s"', container)
    if args.dry:
        logger.info('dry run, skipping ... ')
    else:
        if not args.force and os.path.exists(container):
            raise Exception(
                'Not overwriting existing container "{0}"'.format(container)
            )
        with open(args.container[0], 'wb') as fo:
            frame.mux(
                fo,
                audio_profile=a_prof,
                audio_packets=a_frames,
                video_profile=v_prof,
                video_packets=v_frames,
            )


def parser():
    arg_parser = argparse.ArgumentParser(
        description="""\
libav* front-end extract recorded a/v packets from archives and mux them to
containers.
""",
        version=__version__,
    )
    cmn_parser = argparse.ArgumentParser(add_help=False)
    cmn_parser.add_argument(
        '-l', '--log-level',
        choices=log_levels.keys(),
        default='w',
        help='log level',
    )
    cmd_parsers = arg_parser.add_subparsers(title='commands')
    split_parser(cmd_parsers, [cmn_parser])
    mux_parser(cmd_parsers, [cmn_parser])
    return arg_parser


# Command line parser.
arg_parser = parser()



"""
"""
import dpkt

from . import rtp


def is_rtp_packet(pkt):
    if (not isinstance(pkt.data, dpkt.ip.IP) or
        not isinstance(pkt.data.data, dpkt.udp.UDP)):
        return False
    rtp = dpkt.rtp.RTP(pkt.data.data.data)
    return rtp.version == 2 and (rtp.pt < 64 or 96 <= rtp.pt)


class PCapRTPPacketReader(rtp.RTPPacketReader):
    """
    Iterates and indexes `RTPPacket`s from a PCAP formatted file.
    """
    
    def __init__(self, *args, **kwargs):
        super(PCapRTPPacketReader, self).__init__(*args, **kwargs)
        self.org = self.fo.tell() + dpkt.pcap.FileHdr.__hdr_len__
        self.i = iter(dpkt.pcap.Reader(self.fo))
        try:
            self.i.next()
        except StopIteration:
            pass
        self.reset()

    # rtp.RTPPacketReader

    def __iter__(self):

        def pkts():
            while True:
                _, buf = self.i.next()
                eth = dpkt.ethernet.Ethernet(buf)
                if not is_rtp_packet(eth):
                    continue
                pkt = self.packet_type(eth.data.data.data)
                if not self.packet_filter(pkt):
                    continue
                yield pkt

        self.fo.seek(self.org)
        return pkts()


rtp.RTPPacketReader.register('pcap', PCapRTPPacketReader)



"""
`libav* <http://www.ffmpeg.org/>`_ front-end used to:

- Read stored/archived media (e.g. `pcap <http://www.tcpdump.org/pcap.html>`_ed `rtp <https://tools.ietf.org/html/rfc3550>`_ packets).
- Reconstruct frames (e.g. depacketize a frame from its rtp packets).
- Synchronize reconstructed streams of frame data (e.g. audio and video).
- Mux frame stream(s) to a container (e.g. `mkv <http://www.matroska.org/>`_ file) using libav*.

Motivating use case was:

- depacketizing rtp payloads from `mjr <https://github.com/meetecho/janus-gateway>`_ files and 
- muxing them to an mkv file

which you can do e.g. like:

.. code:: python

    import marm
    
    v_pkts = marm.mjr.MJRRTPPacketReader(
        'path/to/video.mjr', 'rb',
        packet_type=marm.vp8.VP8RTPPacket,
    )
    v_width, v_height = marm.rtp.probe_video_dimensions(v_pkts)
    v_pkts.reset()
    v_frame_rate = marm.rtp.estimate_video_frame_rate(v_pkts, window=10)
    v_pkts.reset()
    v_frames = marm.VideoFrames(v_pkts)
    
    a_pkts = marm.mjr.MJRRTPPacketReader(
        'path/to/audio.mjr', 'rb',
        packet_type=marm.opus.OpusRTPPacket,
    )
    a_frames = marm.Frames(a_pkts)
    
    with open('path/to/muxed.mkv', 'wb') as mkv_fo:
        marm.frame.mux(
            mkv_fo,
            audio_profile={
                'encoder_name': 'libopus',
                'bit_rate': 96000,
                'sample_rate': 48000,
                'time_base': (1, 1000),
            },
            audio_packets=a_frames,
            video_profile={
                'encoder_name': 'libpvx',
                'pix_fmt': marm.VideoFrame.PIX_FMT_YUV420P,
                'width': v_width,
                'height': v_height,
                'frame_rate': v_frame_rate,
                'bit_rate': 4000000,
                'time_base': (1, 1000),
            },
            video_packets=v_frames,
        )

"""
__version__ = '0.1.6'

__all__ = [
    'Frame',
    'Frames',
    'AudioFrame',
    'VideoFrame',
    'VideoFrames',
    'FrameProxy',
    'FrameFilter',
    'frame',
    'rtp',
    'opus',
    'mjr',
    'pcap',
    'FFProbe',
    'FFMPEG',
]

# frames
from .frame import (
    Frame,
    Frames,
    AudioFrame,
    VideoFrame,
    VideoFrames,
    FrameProxy,
    FrameFilter,
)
from . import frame

# packets
from . import rtp
from . import vp8
from . import opus

# stored packets
from . import mjr
from . import pcap

# helper
from .ffmpeg import FFProbe, FFMPEG



import pytest

import marm


@pytest.mark.parametrize(
    ('capture,pt,ssrc,packet_type,expected'), [
        ('streets-of-rage.pcap', 100, 3830765780, marm.vp8.VP8RTPPacket, 1239),
        ('streets-of-rage.pcap', 111, 4286666423, marm.opus.OpusRTPPacket,
         490),
    ])
def test_extract_frames_from_pcap(
        fixtures,
        capture,
        pt,
        ssrc,
        packet_type,
        expected):

    c_path = fixtures.join(capture)
    with c_path.open('rb') as fo:
        pkts = marm.pcap.PCapRTPPacketReader(
            fo,
            packet_type=packet_type,
            packet_filter=lambda pkt: (
                pkt.header.type == pt and
                pkt.header.ssrc == ssrc
            )
        )
        assert sum(1 for _ in pkts) == expected


@pytest.mark.parametrize(
    ('capture,'
     'a_pt,a_ssrc,a_pkt_type,a_enc,'
     'v_pt,v_ssrc,v_pkt_type,v_enc,'
     'fmt,fmt_name,'
     'expected'), [
        ('streets-of-rage.pcap',
         111, 4286666423, marm.opus.OpusRTPPacket, 'libopus',
         100, 3830765780, marm.vp8.VP8RTPPacket, 'libvpx',
         'mkv', 'matroska',
         0),
    ])
def test_mux_frames_from_pcap(
        tmpdir,
        fixtures,
        capture,
        a_pt, a_ssrc, a_pkt_type, a_enc,
        v_pt, v_ssrc, v_pkt_type, v_enc,
        fmt, fmt_name,
        expected):
    c_path = fixtures.join(capture)
    m_path = tmpdir.join('m.{0}'.format(fmt))

    with m_path.open('wb') as m_fo, \
            c_path.open('rb') as a_fo, \
            c_path.open('rb') as v_fo:
        a_pkts = marm.pcap.PCapRTPPacketReader(
            a_fo,
            packet_type=a_pkt_type,
            packet_filter=lambda pkt: (
                pkt.header.type == a_pt and
                pkt.header.ssrc == v_ssrc
            )
        )
        a_pkts = marm.rtp.head_packets(a_pkts, duration=10)

        v_pkts = marm.pcap.PCapRTPPacketReader(
            v_fo,
            packet_type=v_pkt_type,
            packet_filter=lambda pkt: (
                pkt.header.type == v_pt and
                pkt.header.ssrc == v_ssrc
            )
        )
        v_frame_rate = marm.rtp.estimate_video_frame_rate(v_pkts)
        v_pkts.reset()
        v_width, v_height = marm.rtp.probe_video_dimensions(v_pkts)
        v_pkts.reset()
        v_pkts = marm.rtp.head_packets(v_pkts, duration=10)

        marm.frame.mux(
            m_fo,
            audio_profile={
                'encoder_name': a_enc,
                'channel_layout': marm.frame.AudioFrame.CHANNEL_LAYOUT_STEREO,
                'bit_rate': 96000,
                'sample_rate': 48000,
                'time_base': (1, 1000),
            },
            audio_packets=marm.Frames(a_pkts),
            video_profile={
                'encoder_name': v_enc,
                'pix_fmt': marm.VideoFrame.PIX_FMT_YUV420P,
                'width': v_width,
                'height': v_height,
                'frame_rate': v_frame_rate,
                'bit_rate': 4000000,
                'time_base': (1, 1000),
            },
            video_packets=marm.VideoFrames(v_pkts),
        )

    probe = marm.FFProbe([
        '-show_streams',
        '-show_format',
        m_path.strpath,
    ])
    probe()
    assert len(probe.result['streams']) == 2
    assert fmt_name in probe.result['format']['format_name'].split(',')



from __future__ import division

import abc
import collections
import contextlib
import copy
import ctypes
import datetime
import inspect
import itertools
import logging
import math
import os
import StringIO
import struct

from . import ext, VideoFrame


logger = logging.getLogger(__name__)


class RTPHeader(ctypes.BigEndianStructure):
    """
    See https://tools.ietf.org/html/rfc3550#section-5.1
    """

    _pack_ = 1

    _fields_ = [
        ('version', ctypes.c_uint16, 2),
        ('padding', ctypes.c_uint16, 1),
        ('extension', ctypes.c_uint16, 1),
        ('csrccount', ctypes.c_uint16, 4),
        ('markerbit', ctypes.c_uint16, 1),
        ('type', ctypes.c_uint16, 7),
        ('seq_number', ctypes.c_uint16),
        ('timestamp', ctypes.c_uint32),
        ('ssrc', ctypes.c_uint32),
    ]


class RTPTimeMixin(object):
    """
    `RTPPacket` mixin used to compute timing information. 
    """

    # In Hz (e.g. 48000).
    clock_rate = None

    @property
    def ticks(self):
        return self.header.timestamp

    @property
    def secs(self):
        return self.header.timestamp / self.clock_rate

    @property
    def msecs(self):
        return self.secs * 1000


class RTPPayload(object):
    """
    RTP payload interface. It is implemented and associated w/ an `RTPPacket`
    type via `RTPPacket.payload_type`.
    """

    @abc.abstractmethod
    def pack(self, fo=None):
        pass

    @abc.abstractmethod
    def unpack(self, buf):
        pass


class RTPAudioPayloadMixin(object):
    """
    `RTPPayload` mixin used to query audio information.
    """

    @abc.abstractproperty
    def nb_samples(self):
        pass

    @abc.abstractproperty
    def nb_channels(self):
        pass

    @classmethod
    def probe(cls, cur, window=100):
        bit_rate = 96000  # TODO: how to probe/estimate?
        sample_rate = 48000  # TODO: how to probe/estimate?
        with cur.restoring():
            channel_layout = probe_audio_channel_layout(cur)
        return {
            'sample_rate': sample_rate,
            'bit_rate': bit_rate,
            'channel_layout': channel_layout,
        }


class RTPVideoPayloadMixin(object):
    """
    `RTPPayload` mixin used to query video information.
    """

    @abc.abstractproperty
    def is_start_of_frame(self):
        pass

    @abc.abstractproperty
    def is_key_frame(self):
        pass

    @abc.abstractproperty
    def width(self):
        pass

    @abc.abstractproperty
    def height(self):
        pass

    @classmethod
    def probe(cls, cur, window=100):
        with cur.restoring():
            frame_rate = estimate_video_frame_rate(cur, window=window)
        with cur.restoring():
            (width, height) = probe_video_dimensions(cur)
        bit_rate = 4000000  # TODO: estimate?
        pixel_format = VideoFrame.PIX_FMT_YUV420P  # TODO: probe?
        return {
            'pixel_format': pixel_format,
            'frame_rate': frame_rate,
            'bit_rate': bit_rate,
            'width': width,
            'height': height,
        }


class RTPPacket(object):
    """
    Represents an RTP packet. You typically inherit from this an provide a:

    - `type` and
    - `payload_type`

    based on the type of media, e.g. for VP8 video:

    .. code:: python

        class VP8RTPPacket(rtp.RTPTimeMixin, rtp.RTPPacket):

            # rtp.RTPPacket
        
            type = rtp.RTPPacket.VIDEO_TYPE
        
            payload_type = VP8RTPPayload
        
            # RTPTimeMixin
        
            clock_rate = 90000

    """

    AUDIO_TYPE = 'audio'
    VIDEO_TYPE = 'video'

    type = None

    payload_type = None

    def __init__(self, *args, **kwargs):
        if args:
            if kwargs:
                raise TypeError('Mixed positional and keyword arguments')
            elif len(args) > 1:
                raise TypeError('Expected single \'buf\' arg')
            self.unpack(args[0])
        elif kwargs:
            if len(kwargs) == 1 and 'buf' in kwargs:
                self.unpack(kwargs['buf'])
            else:
                self.header = RTPHeader()
                self.csrcs = []
                self.data = ''
                self.pad = 0
                for k, v in kwargs.iteritems():
                    if k not in ('header', 'csrcs', 'data', 'pad'):
                        raise TypeError('Unexpected keyword argument \'{0}\''.format(k))
                    setattr(self, k, v)

    def pack(self, fo=None):
        if fo is None:
            fo, value = StringIO.StringIO(), True
        else:
            value = False
        fo.write(buffer(self.header))
        if self.header.csrccount:
            fo.write(struct.pack('>{0}I'.format(self.header.csrccount), self.csrcs))
        if isinstance(self.data, RTPPayload):
            self.data.pack(fo)
        else:
            fo.write(self.data)
        if self.pad:
            if self.pad - 1:
                fo.write(chr(0) * self.pad - 1)
            fo.write(struct.pack('>B', self.pad))
        if value:
            return fo.getvalue()

    def unpack(self, buf):
        # header
        header = RTPHeader.from_buffer_copy(buf)
        buf = buf[ctypes.sizeof(header):]

        # csrcs
        if header.csrccount:
            fmt = '>{0}I'.format(header.csrccount)
            csrcs = struct.unpack(fmt, buf[:struct.calcsize(fmt)])
            buf = buf[struct.calcsize(fmt):]
        else:
            csrcs = []

        # padding
        if header.padding == 1:
            pad, = struct.unpack('>B', buf[-1])
            logger.debug('stripping %s rtp pad from data', pad)
            buf = buf[:-pad]
        else:
            pad = 0

        # data
        if self.payload_type:
            data = self.payload_type(buf)
        else:
            data = buf

        self.header = header
        self.csrcs = csrcs
        self.data = data
        self.pad = pad


class RTPPacketReader(collections.Iterable):
    """
    Interface for:
    
    - indexing and
    - reading
    
    archived/stored `RTPPacket`s.
    """

    # Default `RTPPacket` type.
    packet_type = RTPPacket

    # Registry for mapping extensions to `RTPPacketReader` implementations.
    formats = {
    }

    @classmethod
    def register(cls, format, type_):
        if not inspect.isclass(type_) or issubclass(RTPPacketReader, type_):
            raise TypeError('type_= must be sub-class of RTPPacketReader')
        if format in cls.formats and cls.formats[format] != type_:
            raise RuntimeError('Different type for "{}" already registered'.format(format))
        cls.formats[format] = type_

    @classmethod
    def open(cls, *args, **kwargs):
        if len(args) == 1 and not isinstance(args[0], basestring):
            src = args[0].name
        else:
            src = args[0]
        _, ext = os.path.splitext(src)
        if ext:
            ext = ext[1:]
        if ext not in cls.formats:
            raise ValueError(
                'No format registered for extension "{0}" from "{1}".'
                .format(ext, src)
            )
        obj = cls.formats[ext](*args, **kwargs)
        return obj

    def __init__(self, *args, **kwargs):
        self.packet_type = kwargs.pop('packet_type', self.packet_type)
        if self.packet_type is None:
            raise Exception('Missing packet_type= and no default for {0}.'.format(self.__name__))
        self.packet_filter = kwargs.pop('packet_filter', None) or (lambda pkt: True)
        if len(args) == 1 and not isinstance(args[0], basestring) and not kwargs:
            self.fo = args[0]
        else:
            self.fo = open(*args, **kwargs)
        self.org = kwargs.pop('org', self.fo.tell())

    def index(self, restore=True):
        """
        Generator yielding positions in file object.
        """
        org = pos = self.fo.tell()
        for _ in self:
            yield pos
            pos = self.fo.tell()
        if restore:
            self.fo.seek(org)

    def reset(self):
        """
        Resets file object to initial packet position.
        """
        self.fo.seek(self.org)

    @contextlib.contextmanager
    def restoring(self):
        pos = self.fo.tell()
        try:
            yield
        finally:
            self.fo.seek(pos)


class RTPCursor(collections.Iterable):
    """
    Cursor used to iterate over a collection or stored `RTPPacket`s.
    """

    def __init__(self, parts, part_type=None, **part_kwargs):
        self.part_type = part_type or RTPPacketReader.open
        self.packet_type = part_kwargs.get('packet_type', RTPPacket)
        self.parts = [
            self._Part(part, self.part_type, part_kwargs) for part in parts
        ]
        self.pos_part, self.pos_pkt = 0, 0
        if self.parts:
            self.part = self.parts[self.pos_part]
        else:
            self.part = None
        self.c = collections.defaultdict(dict)

    def probe(self, window=100):
        return self.packet_type.payload_type.probe(self, window)

    def is_first(self, pos):
        return pos == (0, 0)

    def is_last(self, (part, pkt)):
        if part != len(self.parts) - 1:
            return False
        with self.restoring():
            self.seek((-1, -1))
            return self.tell() == (part, pkt)

    def is_cached(self, tag, key):
        return tag in self.c and key in self.c[tag]

    def cache(self, tag, key, value=None):
        if value is not None:
            self.c[tag][key] = value
            return value
        return self.c[tag].get(key)

    def spans(self, (b_part, b_pkt), (e_part, e_pkt)):
        if e_part == -1:
            e_part = len(self.parts) - 1
        for part in range(b_part, e_part + 1):
            c_b_pkt = b_pkt if part == b_part else 0
            c_e_pkt = e_pkt if part == e_part else -1
            yield part, c_b_pkt, c_e_pkt

    def seek(self, (pos_part, pos_pkt)):
        if self.tell() == (pos_part, pos_pkt):
            return
        if pos_part < 0:
            pos_part = len(self.parts) + pos_part
        if not (0 <= pos_part < len(self.parts)):
            raise IndexError(
                    'Part index {0} out of range [0,{1})'
                    .format(pos_part, len(self.parts))
                )
        part = self.parts[pos_part]
        if not part.is_opened:
            part.open()
        if pos_pkt < 0:
            pos_pkt = len(part) + pos_pkt
        if not (0 <= pos_pkt < len(part)):
            raise IndexError(
                    'Part {0} packet index {1} out of range [0,{2})'
                    .format(part, pos_pkt, len(part))
                )
        if self.part:
            self.part.close()
        self.pos_part, self.pos_pkt = (pos_part, pos_pkt)
        self.part = part

    def tell(self):
        return (self.pos_part, self.pos_pkt)

    def each(self, stop, func):
        for pkt in self.slice(stop):
            func(pkt)

    def count(self, stop, match=None):
        s = {'count': 0}

        def func(pkt):
            if match(pkt):
                s['count'] += 1

        if match is None:
            match = lambda pkt: True
        self.each(stop, func)
        return s['count']

    def compute(self, m, r, stop=None, zero=0, cache=None):
        value = zero
        org = self.tell()
        stop = stop or (-1, -1)
        spans = 0
        for part, b_pos, e_pos in self.spans(org, stop):
            spans += 1
            if cache and self.is_cached(cache, (part, b_pos, e_pos)):
                v = self.cache(cache, (part, b_pos, e_pos))
            else:
                self.seek((part, b_pos))
                md = (m(pkt) for pkt in self.slice(
                    (part, e_pos), inclusive=(e_pos == -1)
                ))
                v = reduce(r, md, zero)
                if cache and (b_pos == 0 and e_pos == -1):
                    self.cache(cache, (part, b_pos, e_pos), v)
            value = reduce(r, [v], value)
        if spans:
            self.seek(stop)
        return value

    def search(self, match, dir='forward'):

        def fwd():
            _, pkt = self._next()
            return match(pkt), pkt

        def bwd():
            _, pkt = self._prev()
            return match(pkt), pkt

        step = {
            'forward': fwd,
            'backward': bwd,
        }.get(dir)
        if not step:
            raise ValueError('Invalid direction "{0}".'.format(dir))

        try:
            while True:
                m, pkt = step()
                if m:
                    return pkt
        except StopIteration:
            pass

    def prev_key_frame(self):
        return self.search(
            lambda pkt: pkt.data.is_start_of_frame and pkt.data.is_key_frame,
            'backward'
        )

    def next_key_frame(self):
        return self.search(
            lambda pkt: pkt.data.is_start_of_frame and pkt.data.is_key_frame,
            'forward'
        )

    def prev_start_of_frame(self):
        return self.search(
            lambda pkt: (
                pkt.data.is_start_of_frame and pkt.data.is_start_of_frame
            ),
            'backward'
        )

    def next_start_of_frame(self):
        return self.search(
            lambda pkt: (
                pkt.data.is_start_of_frame and pkt.data.is_start_of_frame
            ),
            'forward'
        )

    def interval(self, pos=None):
        # FIXME: sum inter-packet delta w/ reset detection?
        if not pos:
            pos = (-1, -1)
        start = self.current().secs
        self.seek(pos)
        delta = self.current().secs - start
        return delta

    def fastforward(self, secs):
        # FIXME: sum inter-packet delta w/ reset detection?
        if not secs:
            return 0
        if secs < 0:
            return self.rewind(-secs)
        start = self.current().secs
        self.search(lambda pkt: pkt.secs - start >= secs, 'forward')
        return (self.current().secs - start) - secs

    def rewind(self, secs):
        # FIXME: sum inter-packet delta w/ reset detection?
        if not secs:
            return 0
        if secs < 0:
            return self.fastforward(-secs)
        start = self.current().secs
        self.search(lambda pkt: start - pkt.secs >= secs, 'backward')
        return (start - self.current().secs) - secs

    def time_cut(self, begin_secs, end_secs, align=True):
        """
        Convert **time** offsets relative to current position in seconds to
        **cursor** positions and optionally align them.

        :param begin_secs: Offset to begin position in seconds from current
            cursor position.

        :param end_secs: Offset to end position in seconds from current cursor
            position.

        :param align: One of:

            - True
            - False
            - "prev"

            If frames span packets (e.g. for video) then alignment moved
            positions to first preceding start of frame packet.

            "prev" is the same as True but **first** moves positions back
            one **before** doing alignment if position is **not** the last one.
            This is useful for getting consistent time cuts when reaching the
            end of the cursor that is later extended with more parts.

        :returns: Tuple of:

            - begin cursor position
            - aligned begin time offset (just `begin_secs` if no `align`)
            - end cursor position
            - aligned end time offset (just `end_secs` if no `align`)

        """
        org = self.tell()

        # head
        b_dt = self.fastforward(begin_secs)
        start = self.tell()

        # tail
        self.seek(org)
        if end_secs is None:
            end_secs, e_dt = self.interval((-1, -1)), 0
        else:
            e_dt = self.fastforward(end_secs)
        stop = self.tell()

        # align
        start_unalign, stop_unalign = start, stop
        if align:
            # framing
            if self.packet_type.type == self.packet_type.VIDEO_TYPE:
                if align == 'prev':
                    if not self.is_first(start):
                        self.prev_to(start)
                        self.prev_start_of_frame()
                        start = self.tell()
                    if not self.is_last(stop):
                        self.prev_to(stop)
                    else:
                        self.seek(stop)
                    self.prev_start_of_frame()
                    stop = self.tell()
                else:
                    self.seek(start)
                    self.prev_start_of_frame()
                    start = self.tell()

                    self.seek(stop)
                    self.prev_start_of_frame()
                    stop = self.tell()
            # no-framing
            else:
                if align == 'prev':
                    if not self.is_first(start):
                        self.prev_to(start)
                        start = self.tell()
                    if not self.is_last(stop):
                        self.prev_to(stop)
                        stop = self.tell()
        self.seek(start_unalign)
        b_align_dt = self.interval(start)
        self.seek(stop_unalign)
        e_align_dt = self.interval(stop)

        # seconds
        start_secs = begin_secs + b_dt + b_align_dt
        stop_secs = end_secs + e_dt + e_align_dt

        return start, start_secs, stop, stop_secs

    def time_positions(self, *args):
        """
        """
        pos = []
        org = self.tell()
        for secs in args:
            self.seek(org)
            self.fastforward(secs)
            pos.append(self.tell())
        return pos

    def trim_frames(self, stop, samples, scale=7, org=(0, 0), zero=0):
        """
        Determines trimming information for a range of packets such that the
        result is evenly divisible by (i.e. **aligns** to) `samples` * number
        of channels. The current position of this cursor is assumed to be the
        start of this range. 
        
        This turns out to be useful when you are e.g. doing **stitched**
        transcoding of audio and need to align on audio frame boundaries which
        *vary* between codecs (e.g. opus 960 -> aac 1024 in samples/channel).
        
        :param stop: Stop position within the cursor.
        :param samples: See `align_frame`.
        :param scale: See `align_frame`.
        :param org: See `align_frame`.
        :param zero: See `align_frame`.

        :return: Tuple of:

            - start position of first packet in this cursor to include
            - stop position of first packet in this cursor to exclude
            - number of samples to **remove** from beginning of decoded
              *source* packets (see e.ffmpeg filter `atrim=start_sample={#}`)
            - slice of encoded *destination* packets (e.g. (1, 5) means keep
              packets 1 - 5)

        """
        start = self.tell()
        size = self.current().data.nb_channels * samples
        target = int(math.ceil(scale / 2))

        b, b_samples, b_off, b_scale = self.align_frame(
            samples, scale=scale, org=org, zero=zero,
        )
        b_idx = 0 if b == start else max(b_scale - target, 0)

        self.seek(stop)
        _, e_samples, e_off, _ = self.align_frame(
            samples, scale=target, org=org, zero=zero,
        )

        total = (e_samples + e_off) - (b_samples + b_off)
        e_idx = int(total / size) - 1

        logger.info(
            'trim %s, %s -> %s, %s w/ offset @ %s, range %s',
            start, stop, b, stop, b_off, (b_idx, e_idx),
        )
        return b, stop, b_off, (b_idx, e_idx)

    def align_frame(self, samples, scale=1, dir='backward', org=(0, 0), zero=0):
        """
        Determines the previous or next location within the cursor where the
        total number of **samples** from some origin (`org`) is evenly
        divisible by (i.e. **aligns** to) `samples` * number of channels.
        
        This turns out to be useful when you are e.g. doing **stitched**
        transcoding of audio and need to align on audio frame boundaries which
        *vary* between codecs (e.g. opus 960 -> aac 1024 in samples/channel).
        
        :param samples: Number of samples per packet.

        :param scale: Attempt to scale alignment this many times. This is
            useful of you want to select e.g. the n-th alignment rather than
            the default first.

        :param dir: Direction to search alignment position.
        
        :param org: Origin relative to which to calculate alignment.
        
        :param zero: Number of samples that preceded `org`. This is useful of
            those packets are *not* actually part of this cursor, such as when
            you do **windowed** packet processing.

        :return: Tuple of:

            - the position of the packet
            - number of samples up to that position 
            - number of samples to add to that to reach **alignment** 
            - the achieved alignment scale factor (<= `scale`)

        """

        def _align(pkt):
            s['count'] += pkt.data.nb_samples * pkt.data.nb_channels
            return s['count'] >= size

        def _scale(pkt):
            s['count'] += pkt.data.nb_samples * pkt.data.nb_channels
            return s['count'] >= scale * size

        start = self.tell()
        pkt = self.current()
        size = pkt.data.nb_channels * samples

        # seek to alignment
        s = {'count': 0}
        pkt = self.search(_align, dir=dir)
        if pkt is None:
            pos, samples, off = start, 0, 0
        else:
            # calculate samples up to alignment
            pos = self.tell()
            self.seek(org)
            samples = self.compute(
                lambda pkt: pkt.data.nb_samples * pkt.data.nb_channels,
                lambda x, y: x + y,
                pos,
                cache='samples',
            ) + zero
            off = 0 if samples % size == 0 else (size - (samples % size))

            # scale
            s = {'count': off}
            self.search(_scale, dir=dir)
            scaled_samples = samples - (s['count'] - off)
            scaled_off = 0 if scaled_samples % size == 0 else size - (scaled_samples % size)
            scaled_pos = self.tell()
            scale = int(((samples + off) - (scaled_samples + scaled_off)) / size)
            pos, samples, off = scaled_pos, scaled_samples, scaled_off

        logger.info(
            'aligned %s -> %s @ %s samples w/ trim %s',
            start, pos, samples, off
        )
        return pos, samples, off, scale

    def prev_to(self, pos):
        self.seek(pos)
        return self.prev()

    def prev(self):
        _, pkt = self._prev()
        return pkt

    def slice(self, stop=None, inclusive=False):
        if stop is None:
            stop = self.tell()[0], -1
        if stop[1] == -1:
            with self.restoring():
                self.seek(stop)
                stop = self.tell()
            inclusive = True
        if stop < self.tell():
            yield self.current()
            try:
                while True:
                    pos, pkt = self._prev()
                    if pos <= stop:
                        if inclusive:
                            yield pkt
                        break
                    yield pkt
            except StopIteration:
                pass
        elif stop > self.tell():
            yield self.current()
            try:
                while True:
                    pos, pkt = self._next()
                    if pos >= stop:
                        if inclusive:
                            yield pkt
                        break
                    yield pkt
            except StopIteration:
                pass

    def time_slice(self, begin_secs, end_secs, align=True):
        org = self.tell()

        # start
        self.fastforward(begin_secs)
        begin = self.tell()
        if align and self.packet_type.type == self.packet_type.VIDEO_TYPE:
            self.next_key_frame()
        start = self.tell()
        self.seek(begin)
        start_secs = begin_secs + self.interval(start)

        # stop
        if end_secs is None:
            self.seek(org)
            end_secs = self.interval((-1, -1))
        self.seek(org)
        self.fastforward(end_secs)
        end = self.tell()
        if align and self.packet_type.type == self.packet_type.VIDEO_TYPE:
            self.prev_start_of_frame()
        stop = self.tell()
        stop_secs = end_secs + self.interval(end)

        # iterator

        class Slice(collections.Iterator):

            def __init__(self, cur, start, stop):
                self.cur, self.start, self.stop = cur, start, stop
                self.i = None

            # collections.Iterator

            def __iter__(self):
                self.cur.seek(self.start)
                self.i = self.cur.slice(self.stop)
                return self

            def next(self):
                return self.i.next()

        return start_secs, stop_secs, Slice(self, start, stop)

    def current(self):
        if not self.part.is_opened:
            self.part.open()
        return self.part.packet(self.pos_pkt)

    def copy(self):
        return copy.copy(self)

    def __copy__(self):
        obj = type(self)(
            part_type=self.part_type,
            parts=[part.file for part in self.parts],
        )
        obj.seek(self.tell())
        return obj

    @contextlib.contextmanager
    def restoring(self):
        pos = self.tell()
        try:
            yield
        finally:
            self.seek(pos)

    # collections.Iterable

    def __iter__(self):
        if self.part is None:
            return
        try:
            yield self.current()
        except IndexError:
            raise StopIteration()
        while True:
            _, pkt = self._next()
            yield pkt

    # internals

    class _Part(collections.Sequence):

        def __init__(self, file, part_type, part_kwargs):
            self.file = file
            self.part_type = part_type
            self.part_kwargs = part_kwargs
            self.pkts = None
            self.i = None
            self.idx = []

        def open(self):
            self.close()
            self.pkts = self.part_type(self.file, **self.part_kwargs)
            for pos in self.pkts.index():
                self.idx.append(pos)
            self.i = iter(self.pkts)

        def close(self):
            self.pkts = None
            self.i = None
            del self.idx[:]

        @property
        def name(self):
            return (
                self.file
                if isinstance(self.file, basestring)
                else getattr(self.file, 'name', '<memory>')
            )

        @property
        def is_opened(self):
            return self.pkts is not None

        @property
        def is_closed(self):
            return not self.is_opened

        def packet(self, i):
            if self.is_closed:
                self.open()
            self.pkts.fo.seek(self.idx[i], os.SEEK_SET)
            return self.i.next()

        # collections.Sequence

        def __getitem__(self, index):
            return self.parts[index]

        def __len__(self):
            return len(self.idx)

    def _next(self):
        # next
        (pos_part, pos_pkt) = (self.pos_part, self.pos_pkt)
        try:
            self.pos_pkt += 1
            if not (0 <= self.pos_pkt < len(self.part)):
                if self.pos_part + 1 < len(self.parts):
                    self.pos_part += 1
                    self.pos_pkt = 0
                    self.part = self.parts[self.pos_part]
            if self.part is None:
                raise StopIteration
            if not self.part.is_opened:
                self.part.open()
            if not (0 <= self.pos_part < len(self.parts)):
                raise StopIteration
            if not (0 <= self.pos_pkt < len(self.part)):
                raise StopIteration
        except StopIteration:
            (self.pos_part, self.pos_pkt) = (pos_part, pos_pkt)
            raise

        # read
        pos, pkt = self.tell(), self.part.packet(self.pos_pkt)
        return pos, pkt

    def _prev(self):
        # prev
        (pos_part, pos_pkt) = (self.pos_part, self.pos_pkt)
        try:
            self.pos_pkt -= 1
            while self.pos_pkt < 0:
                self.pos_part -= 1
                if self.pos_part < 0:
                    self.pos_part, self.pos_pkt = 0, 0
                    self.part = self.parts[self.pos_part]
                    raise StopIteration
                self.part = self.parts[self.pos_part]
                if not self.part.is_opened:
                    self.part.open()
                self.pos_pkt = len(self.part) - 1
            if self.part is None:
                raise StopIteration
            if not self.part.is_opened:
                self.part.open()
        except StopIteration:
            (self.pos_part, self.pos_pkt) = (pos_part, pos_pkt)
            raise

        # read
        pos, pkt = self.tell(), self.part.packet(self.pos_pkt)
        return pos, pkt


def head_packets(packets, count=None, duration=None):
    """
    Iterator for first n packets where n is capped by a:
    
    - packet count and/or
    - duration in seconds
    
    """
    s = {
        'epoch': None,
        'count': 0
    }

    def _duration(pkt):
        if s['epoch'] is None:
            s['epoch'] = pkt.secs
        return pkt.secs - s['epoch'] < duration

    def _count(pkt):
        s['count'] += 1
        return s['count'] < count

    ps = []
    if duration is not None:
        ps.append(_duration)
    if count is not None:
        ps.append(_count)
    predicate = lambda pkt: all(p(pkt) for p in ps)

    return itertools.takewhile(predicate, packets)


def probe_video_dimensions(packets):
    """
    Finds first start-of-frame packet and extracts video width and height from
    it.
    """
    for pkt in packets:
        if pkt.data.is_start_of_frame and pkt.data.is_key_frame:
            return pkt.data.width, pkt.data.height


def estimate_video_frame_rate(packets, window=10, min_window=10):
    """
    Finds `window` start-of-frame packets and uses their timestamps to estimate
    video frame rate. 
    """
    ts = []
    for pkt in packets:
        if pkt.data.is_start_of_frame:
            ts.append(pkt.secs)
            if len(ts) >= window:
                break
    if len(ts) < min_window:
        raise ValueError(
            'Not enough start-of-frame packets %s (< %s).'
            .format(len(ts), min_window)
        )
    return (len(ts) - 1) / (ts[-1] - ts[0])


def probe_audio_channel_layout(packets):
    """
    Determines audio channel layout from first packet.
    """
    pkt = iter(packets).next()
    nb_channels = pkt.data.nb_channels
    if nb_channels == 1:
        return ext.AV_CH_LAYOUT_MONO
    elif nb_channels == 2:
        return ext.AV_CH_LAYOUT_STEREO
    raise ValueError('Unsupported number of channel {0}.'.format(nb_channels))


def split_packets(packets, duration=None, count=None):
    """
    Splits packets into n-sized packet chunks where n is capped by a:
    
    - packet count and/or
    - duration in seconds or `datetime.timedelta`
    
    """
    packets = iter(packets)

    # slice

    s = {
        'epoch': None,
        'count': 0,
        'last': None
    }

    def _duration(pkt):
        if s['epoch'] is None:
            s['epoch'] = pkt.secs
        return pkt.secs - s['epoch'] < duration

    def _count(pkt):
        s['count'] += 1
        return s['count'] < count

    ps = []
    if duration is not None:
        if isinstance(duration, datetime.timedelta):
            duration = duration.total_seconds()
        ps.append(_duration)
    if count is not None:
        ps.append(_count)
    _predicate = lambda pkt: all(p(pkt) for p in ps)

    def _slice():
        if s['last']:
            pkt, s['last'] = s['last'], None
        else:
            pkt = packets.next()
        if not _predicate(pkt):
            s['last'] = pkt
            return
        yield pkt
        while True:
            pkt = packets.next()
            if not _predicate(pkt):
                s['last'] = pkt
                break
            yield pkt

    # slices

    while True:
        # reset state
        s['epoch'] = None
        s['count'] = 0

        # slice
        pkts = _slice()
        try:
            pkt = pkts.next()
        except StopIteration:
            # nothing so we're done
            break
        yield itertools.chain([pkt], pkts)



import itertools

import pytest

import marm


@pytest.mark.parametrize(
    'mpegts,ccs', [
        ('sonic.ts', {0: 4, 17: 9, 4096: 4, 256: 2, 257: 8})
    ]
)
def test_mpegts_last_ccs(fixtures, mpegts, ccs):
    with fixtures.join(mpegts).open('rb') as fo:
        assert marm.frame.last_mpegts_ccs(fo) == ccs


@pytest.mark.parametrize(
    'v_mjr,v_pkt_type,a_mjr,a_pkt_type,time_slice,time_cuts', [
        ('sonic-v.mjr', marm.vp8.VP8RTPPacket,
         'sonic-a.mjr', marm.opus.OpusRTPPacket,
         (0, 60),
         [11, 12, 12, 13]),
        ('sonic-v.mjr', marm.vp8.VP8RTPPacket,
         'sonic-a.mjr', marm.opus.OpusRTPPacket,
         (0, 15),
         [7.5]),
    ]
)
def test_mpegts_stitch_one(
        fixtures,
        tmpdir,
        v_mjr, v_pkt_type,
        a_mjr, a_pkt_type,
        time_slice,
        time_cuts):
    # time slice mjr
    v_mjr = pytest.time_slice(
        tmpdir.join('v.mjr'),
        fixtures.join(v_mjr), v_pkt_type, time_slice,
    )
    a_mjr = pytest.time_slice(
        tmpdir.join('a.mjr'),
        fixtures.join(a_mjr), a_pkt_type, time_slice,
    )

    # full
    v_cuts, a_cuts = pytest.time_cut(
        v_mjr, v_pkt_type,
        a_mjr, a_pkt_type, 1024,  # *fixed* @ 1024 for aac
    )
    mkv, v_drop, a_trim, a_range = pytest.mux_time_cuts(
        tmpdir, 'f.mkv',
        v_mjr, v_pkt_type, a_mjr, a_pkt_type,
        zip(v_cuts, a_cuts),
    )[0]
    f_ts_p = tmpdir.join('{0}-p.ts'.format(mkv.purebasename))
    ffmpeg = marm.FFMPEG([
        '-y',
        '-i', mkv.strpath,
        '-c:v', 'h264',
        '-filter:v', 'select=gte(n\,{0})'.format(v_drop),
        '-c:a', 'aac', '-strict', '-2',
        '-b:a', '22k',
        '-r:a', '48k',
        '-filter:a', 'atrim=start_sample={0}'.format(a_trim),
        '-copyts',
        '-mpegts_copyts', '1',
        '-avoid_negative_ts', '0',
        f_ts_p.strpath,
    ])
    ffmpeg()
    f_ts = tmpdir.join('f.ts')
    with f_ts_p.open('rb') as sfo, f_ts.open('wb') as dfo:
        marm.frame.remux(
            dfo, sfo,
            marm.FrameFilter.range(1, *a_range).shift(1),
            options=[
                ('copyts', '1'),
                ('mpegts_copyts', '1'),
                ('avoid_negative_ts', '0'),
            ],
        )

    # calculate mjr "cuts"
    v_cuts, a_cuts = pytest.time_cut(
        v_mjr, v_pkt_type,
        a_mjr, a_pkt_type, 1024,  # *fixed* @ 1024 for aac
        * time_cuts
    )

    # mux each "cut" to mkv
    mkvs = pytest.mux_time_cuts(
        tmpdir, 'c-{i}.mkv',
        v_mjr, v_pkt_type, a_mjr, a_pkt_type,
        zip(v_cuts, a_cuts),
    )

    # xcode "cut" mkv to mpegts
    tss = []
    for mkv, v_drop, a_trim, a_range in mkvs:
        # xcode mkv -> mpegt(v=h264, a=aac)
        ts_p = tmpdir.join('{0}-p.ts'.format(mkv.purebasename))
        ffmpeg = marm.FFMPEG([
            '-y',
            '-i', mkv.strpath,
            '-c:v', 'h264',
            '-filter:v', 'select=gte(n\,{0})'.format(v_drop),
            '-c:a', 'aac', '-strict', '-2',
            '-b:a', '22k',
            '-r:a', '48k',
            '-filter:a', 'atrim=start_sample={0}'.format(a_trim),
            '-copyts',
            '-mpegts_copyts', '1',
            '-avoid_negative_ts', '0',
            ts_p.strpath,
        ])
        ffmpeg()

        # remux mpegts to:
        # - force mpegts continuity counts across "cuts" and
        # - drop aac initial padding (aka audio codec priming samples)
        ts = tmpdir.join('{0}.ts'.format(mkv.purebasename))
        with ts_p.open('rb') as sfo, ts.open('wb') as dfo:
            marm.frame.remux(
                dfo, sfo,
                marm.FrameFilter.range(1, *a_range).shift(1),
                options=[
                    ('copyts', '1'),
                    ('mpegts_copyts', '1'),
                    ('avoid_negative_ts', '0'),
                ],
            )

        tss.append(ts)

    # and concat them
    c_ts = tmpdir.join('c.ts')
    with c_ts.open('wb') as dfo:
        for ts in tss:
            with ts.open('rb') as sfo:
                dfo.write(sfo.read())

    # compare timing
    c_pkts = pytest.packets(c_ts, bucket='stream', parse=lambda p: p['pts'])
    f_pkts = pytest.packets(f_ts, bucket='stream', parse=lambda p: p['pts'])
    assert len(c_pkts) == len(f_pkts)
    for c_p, f_p in itertools.izip(c_pkts.itervalues(), f_pkts.itervalues()):
        assert len(c_p) == len(f_p)
        assert min(c_p) == min(f_p)
        assert max(c_p) == max(f_p)
        assert sum(
            c_pts != f_pts
            for c_pts, f_pts in itertools.izip(sorted(c_p), sorted(f_p))
        ) < 0.01 * len(c_p)


@pytest.mark.parametrize(
    'v_mjr,v_pkt_type,a_mjr,a_pkt_type,time_slice,time_cuts', [
        ('sonic-v.mjr', marm.vp8.VP8RTPPacket,
         'sonic-a.mjr', marm.opus.OpusRTPPacket,
         (0, 20),
         [5, 6, 7]),
    ]
)
def test_mpegts_stitch_many(
        fixtures,
        tmpdir,
        v_mjr, v_pkt_type,
        a_mjr, a_pkt_type,
        time_slice,
        time_cuts):
    # slice and split mjr
    v_mjr = pytest.time_slice(
        tmpdir.join('v.mjr'),
        fixtures.join(v_mjr), v_pkt_type, time_slice,
        sort=True,
    )
    a_mjr = pytest.time_slice(
        tmpdir.join('a.mjr'),
        fixtures.join(a_mjr), a_pkt_type, time_slice,
        sort=True,
    )
    v_mjrs, a_mjrs = pytest.time_split(
        tmpdir, v_mjr, v_pkt_type, a_mjr, a_pkt_type, *time_cuts
    )

    # full
    v_cuts, a_cuts = pytest.time_cut(
        v_mjrs, v_pkt_type,
        a_mjrs, a_pkt_type, 1024,  # *fixed* @ 1024 for aac
    )
    mkv, v_drop, a_trim, a_range = pytest.mux_time_cuts(
        tmpdir, 'f.mkv',
        v_mjrs, v_pkt_type, a_mjrs, a_pkt_type,
        zip(v_cuts, a_cuts),
    )[0]
    f_ts_p = tmpdir.join('{0}-p.ts'.format(mkv.purebasename))
    ffmpeg = marm.FFMPEG([
        '-y',
        '-i', mkv.strpath,
        '-c:v', 'h264',
        '-filter:v', 'select=gte(n\,{0})'.format(v_drop),
        '-c:a', 'aac', '-strict', '-2',
        '-b:a', '22k',
        '-r:a', '48k',
        '-filter:a', 'atrim=start_sample={0}'.format(a_trim),
        '-copyts',
        '-mpegts_copyts', '1',
        '-avoid_negative_ts', '0',
        f_ts_p.strpath,
    ])
    ffmpeg()
    f_ts = tmpdir.join('f.ts')
    with f_ts_p.open('rb') as sfo, f_ts.open('wb') as dfo:
        marm.frame.remux(
            dfo, sfo,
            marm.FrameFilter.range(1, a_range[0], a_range[1] + 1),
            options=[
                ('copyts', '1'),
                ('mpegts_copyts', '1'),
                ('avoid_negative_ts', '0'),
            ],
        )

    # calculate cuts
    v_cuts, a_cuts = pytest.time_cut(
        v_mjrs, v_pkt_type,
        a_mjrs, a_pkt_type, 1024,  # *fixed* @ 1024 for aac
        * time_cuts
    )

    # mux each cut to mkv
    mkvs = pytest.mux_time_cuts(
        tmpdir, 'c-{i}.mkv',
        v_mjrs, v_pkt_type,
        a_mjrs, a_pkt_type,
        zip(v_cuts, a_cuts),
    )

    # xcode cut mkv to mpegts
    tss = []
    for i, (mkv, v_drop, a_trim, a_range) in enumerate(mkvs):
        # xcode mkv -> mpegt(v=h264, a=aac)
        ts_p = tmpdir.join('{0}-p.ts'.format(mkv.purebasename))
        ffmpeg = marm.FFMPEG([
            '-y',
            '-i', mkv.strpath,
            '-c:v', 'h264',
            '-filter:v', 'select=gte(n\,{0})'.format(v_drop),
            '-c:a', 'aac', '-strict', '-2',
            '-b:a', '22k',
            '-r:a', '48k',
            '-filter:a', 'atrim=start_sample={0}'.format(a_trim),
            '-copyts',
            '-mpegts_copyts', '1',
            '-avoid_negative_ts', '0',
            ts_p.strpath,
        ])
        ffmpeg()

        # remux mpegts to:
        # - force mpegts continuity counts across "cuts" and
        # - drop aac initial padding (aka audio codec priming samples)
        ts = tmpdir.join('{0}.ts'.format(mkv.purebasename))
        if i == 0:
            a_range = a_range[0], a_range[1] + 1
        else:
            a_range = a_range[0] + 1, a_range[1] + 1
        with ts.open('wb') as dfo, ts_p.open('rb') as sfo:
            marm.frame.remux(
                dfo, sfo,
                marm.FrameFilter.range(1, *a_range),
                options=[
                    ('copyts', '1'),
                    ('mpegts_copyts', '1'),
                    ('avoid_negative_ts', '0'),
                ],
            )

        tss.append(ts)

    # and concat them
    c_ts = tmpdir.join('c.ts')
    with c_ts.open('wb') as dfo:
        for ts in tss:
            with ts.open('rb') as sfo:
                dfo.write(sfo.read())

    # compare counts and timing
    c_pkts = pytest.packets(c_ts, bucket='stream', parse=lambda p: p['pts'])
    f_pkts = pytest.packets(f_ts, bucket='stream', parse=lambda p: p['pts'])
    assert len(c_pkts) == len(f_pkts)
    for c_p, f_p in itertools.izip(c_pkts.itervalues(), f_pkts.itervalues()):
        assert len(c_p) == len(f_p)
        assert min(c_p) == min(f_p)
        assert max(c_p) == max(f_p)
        assert sum(
            c_pts != f_pts
            for c_pts, f_pts in itertools.izip(sorted(c_p), sorted(f_p))
        ) < 0.01 * len(c_p)


@pytest.mark.parametrize(
    'v_mjr,v_pkt_type,a_mjr,a_pkt_type,time_slice,time_cuts,window', [
        ('sonic-v.mjr', marm.vp8.VP8RTPPacket,
         'sonic-a.mjr', marm.opus.OpusRTPPacket,
         (10, 70),
         [5] * 10,
         (4, 1, 0)),
        ('sonic-v.mjr', marm.vp8.VP8RTPPacket,
         'sonic-a.mjr', marm.opus.OpusRTPPacket,
         (10, 70),
         [6.5] * 7,
         (3, 2, 0)),
    ]
)
def test_mpegts_stitch_window(
        tmpdir,
        fixtures,
        v_mjr, v_pkt_type,
        a_mjr, a_pkt_type,
        time_slice,
        time_cuts,
        window):
    a_idx = 1

    # slice mjrs
    v_mjr = pytest.time_slice(
        tmpdir.join('v.mjr'),
        fixtures.join(v_mjr), v_pkt_type, time_slice,
        sort=True,
    )
    a_mjr = pytest.time_slice(
        tmpdir.join('a.mjr'),
        fixtures.join(a_mjr), a_pkt_type, time_slice,
        sort=True,
    )

    # split mjrs
    v_mjrs, a_mjrs = pytest.time_split(
        tmpdir,
        v_mjr, v_pkt_type,
        a_mjr, a_pkt_type,
        *time_cuts
    )

    # full mpegts (we'll compare stitched to this)
    v_cuts, a_cuts = pytest.time_cut(
        v_mjrs, v_pkt_type,
        a_mjrs, a_pkt_type, 1024,  # *fixed* @ 1024 for aac
    )
    mkv, v_drop, a_trim, a_range = pytest.mux_time_cuts(
        tmpdir, 'f.mkv',
        v_mjrs, v_pkt_type, a_mjrs, a_pkt_type,
        zip(v_cuts, a_cuts),
    )[0]
    f_ts_p = tmpdir.join('{0}-p.ts'.format(mkv.purebasename))
    ffmpeg = marm.FFMPEG([
        '-y',
        '-i', mkv.strpath,
        '-c:v', 'h264',
        '-filter:v', 'select=gte(n\,{0})'.format(v_drop),
        '-c:a', 'aac', '-strict', '-2',
        '-b:a', '22k',
        '-r:a', '48k',
        '-filter:a', 'atrim=start_sample={0}'.format(a_trim),
        '-copyts',
        '-mpegts_copyts', '1',
        '-avoid_negative_ts', '0',
        f_ts_p.strpath,
    ])
    ffmpeg()
    f_ts = tmpdir.join('f.ts')
    with f_ts_p.open('rb') as sfo, f_ts.open('wb') as dfo:
        marm.frame.remux(
            dfo, sfo,
            marm.FrameFilter.range(a_idx, a_range[0], a_range[1] + 1),
            options=[
                ('copyts', '1'),
                ('mpegts_copyts', '1'),
                ('avoid_negative_ts', '0'),
            ],
        )

    # probe
    v_prof = pytest.probe(
        marm.rtp.RTPCursor([v_mjrs[0].strpath], packet_type=v_pkt_type)
    )
    a_prof = pytest.probe(
        marm.rtp.RTPCursor([a_mjrs[0].strpath], packet_type=a_pkt_type)
    )

    # xcode windows
    lead, size, lag = window
    total = len(v_mjrs)
    for j, i in enumerate(range(0, total, size)):
        l, u = max(0, i - lead), min(total, i + size - 1 + lag)

        v_wnd = [mjr.strpath for mjr in v_mjrs[l:u + 1]]
        a_wnd = [mjr.strpath for mjr in a_mjrs[l:u + 1]]
        b = i - l
        e = min(b + size, len(v_wnd))

        # v window
        v_cur = marm.rtp.RTPCursor(v_wnd, packet_type=v_pkt_type)
        v_cur.seek((b, 0))
        with v_cur.restoring():
            v_delta = v_cur.interval((e, 0)) if e != len(v_wnd) else None
        v_start, _, v_stop, _ = v_cur.time_cut(0, v_delta, align='prev')
        v_cur.seek(v_start)
        if v_cur.current().data.is_key_frame:
            v_key = v_start
        else:
            v_cur.prev_key_frame()
            v_key = v_cur.tell()
        v_cur.seek(v_key)
        v_drop = v_cur.count(v_start, lambda pkt: pkt.data.is_start_of_frame)

        # a window
        a_cur = marm.rtp.RTPCursor(a_wnd, packet_type=a_pkt_type)
        a_cur.seek((b, 0))
        with a_cur.restoring():
            a_delta = a_cur.interval((e, 0)) if e != len(a_wnd) else None
        a_start, _, a_stop, _ = a_cur.time_cut(0, a_delta, align='prev')
        a_cur.seek(a_start)
        a_start, a_stop, a_trim, a_range = a_cur.trim_frames(
            a_stop, 1024, zero=marm.rtp.RTPCursor(
                [mjr.strpath for mjr in a_mjrs[:l]],
                packet_type=marm.opus.OpusRTPPacket,
            ).compute(
                lambda pkt: pkt.data.nb_samples * pkt.data.nb_channels,
                lambda x, y: x + y,
            )
        )

        # frames
        v_cur.seek(v_key)
        assert v_cur.tell() == v_key
        assert v_cur.current().data.is_key_frame
        v_frames = marm.VideoFrames(v_cur.slice(v_stop), -v_prof['msec_org'])
        a_cur.seek(a_start)
        assert a_cur.tell() == a_start
        a_frames = marm.Frames(a_cur.slice(a_stop), -a_prof['msec_org'])

        # mux window
        mkv = tmpdir.join('c-{0:03}.mkv'.format(j))
        with mkv.open('wb') as fo:
            marm.frame.mux(
                fo,
                video_profile={
                    'encoder_name': 'libvpx',
                    'pix_fmt': marm.VideoFrame.PIX_FMT_YUV420P,
                    'width': v_prof['width'],
                    'height': v_prof['height'],
                    'frame_rate': v_prof['frame_rate'],
                    'bit_rate': v_prof['bit_rate'],
                    'time_base': (1, 1000),
                },
                video_packets=v_frames,
                audio_profile={
                    'encoder_name': 'libopus',
                    'bit_rate': a_prof['bit_rate'],
                    'sample_rate': a_prof['sample_rate'],
                    'channel_layout': a_prof['channel_layout'],
                    'time_base': (1, 1000),
                    'initial_padding': 0,
                },
                audio_packets=a_frames,
            )
        assert v_cur.tell() == v_stop
        assert v_cur.current().data.is_start_of_frame
        assert a_cur.tell() == a_stop

        # xcode window
        p_ts = tmpdir.join('p-{0:03}.ts'.format(j))
        ffmpeg = marm.FFMPEG([
            '-y',
            '-i', mkv.strpath,
            '-c:v', 'h264',
            '-filter:v', 'select=gte(n\,{0})'.format(v_drop),
            '-c:a', 'aac', '-strict', '-2',
            '-b:a', '22k',
            '-r:a', '48k',
            '-filter:a', 'atrim=start_sample={0}'.format(a_trim),
            '-copyts',
            '-mpegts_copyts', '1',
            '-avoid_negative_ts', '0',
            p_ts.strpath,
        ])
        ffmpeg()
        ts = tmpdir.join('c-{0:03}.ts'.format(j))
        with ts.open('wb') as dfo, p_ts.open('rb') as sfo:
            if b == 0:
                # keep codec priming packet
                a_range = (a_range[0], a_range[1] + 1)
            else:
                # drop codec priming packet
                a_range = (a_range[0] + 1, a_range[1] + 1)
            marm.frame.remux(
                dfo, sfo,
                marm.FrameFilter.range(a_idx, *a_range),
                options=[
                    ('copyts', '1'),
                    ('mpegts_copyts', '1'),
                    ('avoid_negative_ts', '0'),
                ],
            )

    # concat widows
    c_ts = tmpdir.join('c.ts')
    with c_ts.open('wb') as dfo:
        for ts in sorted(tmpdir.listdir('c-*.ts')):
            with ts.open('rb') as sfo:
                dfo.write(sfo.read())

    # compare to full
    c_pkts = pytest.packets(c_ts, bucket='stream', parse=lambda p: p['pts'])
    f_pkts = pytest.packets(f_ts, bucket='stream', parse=lambda p: p['pts'])
    assert len(c_pkts) == len(f_pkts)
    for c_p, f_p in itertools.izip(c_pkts.itervalues(), f_pkts.itervalues()):
        assert len(c_p) == len(f_p)
        assert min(c_p) == min(f_p)
        assert max(c_p) == max(f_p)


@pytest.mark.parametrize(
    'v_store,v_pkt,v_enc,a_store,a_pkt,a_enc,time_slice,interval,delta', [
        ('sonic-v.mjr', marm.vp8.VP8RTPPacket, 'libvpx',
         'sonic-a.mjr', marm.opus.OpusRTPPacket, 'libopus',
         (0, 30),
         5,
         1.0),
    ]
)
def test_mpegts_segment(
        fixtures,
        tmpdir,
        v_store, v_pkt, v_enc,
        a_store, a_pkt, a_enc,
        time_slice,
        interval,
        delta):
    # time slice src
    v_src = pytest.time_slice(
        tmpdir.join('v.mjr'),
        fixtures.join(v_store), v_pkt, time_slice,
    )
    a_src = pytest.time_slice(
        tmpdir.join('a.mjr'),
        fixtures.join(a_store), a_pkt, time_slice,
    )

    # mux src
    src = tmpdir.join('c.mkv')
    pytest.mux(
        src,
        v_src, v_pkt, v_enc,
        a_src, a_pkt, a_enc,
    )

    # xcode src -> full
    full = tmpdir.join('f.ts')
    ffmpeg = marm.FFMPEG([
        '-y',
        '-i', src.strpath,
        '-c:v', 'h264',
        '-r', '25',
        '-force_key_frames', 'expr:gte(t,n_forced*{0})'.format(5),
        '-c:a', 'aac', '-strict', '-2',
        '-b:a', '22k',
        '-r:a', '48k',
        '-copyts',
        '-mpegts_copyts', '1',
        '-avoid_negative_ts', '0',
        full.strpath,
    ])
    ffmpeg()

    # split full -> segments
    seg_fmt = tmpdir.join('s-%03d.ts')
    ccs = {}
    with full.open('rb') as sfo:
        marm.frame.segment(
            seg_fmt.strpath,
            'mpegts',
            sfo,
            time=interval,
            time_delta=1 / (1 * 25.0),
            mpegts_ccs=ccs,
            options=[
                ('copyts', '1'),
                ('mpegts_copyts', '1'),
                ('avoid_negative_ts', '0'),
            ],
        )

    # check segment packets
    for s in sorted(tmpdir.listdir('s-*.ts')):
        s_pkts = pytest.packets(s, bucket='stream')

        # videos start w/ keyframe
        assert all(
            s_pkts[idx][0]['flags'] == 'K'
            for idx in s_pkts.keys() if s_pkts[idx][0]['codec_type'] == 'video'
        )

        # durations are close to interval
        assert all(
            abs(interval - (
                float(pkts[-1]['pts_time']) - float(pkts[0]['pts_time'])
            )) < delta
            for pkts in s_pkts.values()[:-1]
        )

    # concat segments
    concat = tmpdir.join('c.ts')
    with concat.open('wb') as dfo:
        for rseg in sorted(tmpdir.listdir('s-*.ts')):
            with rseg.open('rb') as sfo:
                dfo.write(sfo.read())

    # and compare to full
    full_pkts = pytest.packets(
        full, bucket='stream', parse=lambda p: p['pts']
    )
    concat_pkts = pytest.packets(
        concat, bucket='stream', parse=lambda p: p['pts']
    )
    assert full_pkts == concat_pkts



import pytest

import marm.cli


@pytest.mark.parametrize(
    ('stored,pkt_type,pkt_filter,dur,splits,packets'), [
        ('sonic-v.mjr', 'vp8', None, 10.0, 12, 9654),
        ('sonic-a.mjr', 'opus', None, 10.0, 12, 5996),
        ('streets-of-rage.pcap', 'vp8', 'ssrc=3830765780', 2.0, 5, 1239),
        ('streets-of-rage.pcap', 'opus', 'ssrc=4286666423', 2.0, 5, 490),
    ]
)
def test_cli_split(
        tmpdir,
        fixtures,
        stored,
        pkt_type,
        pkt_filter,
        dur,
        splits,
        packets):
    src_path = fixtures.join(stored)

    args = ['split', pkt_type, src_path, tmpdir, '--dur', dur]
    if pkt_filter:
        args.extend(['--filter', pkt_filter])
    parsed = marm.cli.arg_parser.parse_args(map(str, args))
    parsed.cmd(parsed)

    pattern = '{0}*{1}'.format(src_path.purebasename, 'mjr')
    split_paths = list(tmpdir.visit(pattern))

    assert len(split_paths) == splits
    assert sum(
            1
            for split_path in split_paths
            for _ in marm.rtp.RTPPacketReader.open(str(split_path))
        ) == packets


@pytest.mark.parametrize(
    ('a_stored,a_type,a_filter,a_dur,v_stored,v_type,v_filter,v_dur,muxed'), [
        ('sonic-a.mjr', 'opus', None, 10.0,
         'sonic-v.mjr', 'vp8', None, 10.0,
         'mux.mkv'),
        ('streets-of-rage.pcap', 'opus', 'ssrc=4286666423', 2.0,
         'streets-of-rage.pcap', 'vp8', 'ssrc=3830765780', 2.0,
         'mux.mkv'),
    ]
)
def test_cli_mux(
        tmpdir,
        fixtures,
        a_stored, a_type, a_filter, a_dur,
        v_stored, v_type, v_filter, v_dur,
        muxed):
    dst = tmpdir.join(muxed)

    # split a
    a_src = fixtures.join(a_stored)
    a_splits = tmpdir.join('a-{0}.{1}'.format('{part:02}', 'mjr'))
    args = ['split', a_type, a_src, a_splits, '--dur', a_dur]
    if a_filter:
        args.extend(['--filter', a_filter])
    parsed = marm.cli.arg_parser.parse_args(map(str, args))
    parsed.cmd(parsed)
    a_splits = sorted(tmpdir.visit('a-*.{0}'.format('mjr')))

    # split v
    v_src = fixtures.join(v_stored)
    v_splits = tmpdir.join('v-{0}.{1}'.format('{part:02}', 'mjr'))
    args = ['split', v_type, v_src, v_splits, '--dur', v_dur]
    if v_filter:
        args.extend(['--filter', v_filter])
    parsed = marm.cli.arg_parser.parse_args(map(str, args))
    parsed.cmd(parsed)
    v_splits = sorted(tmpdir.visit('v-*.{0}'.format('mjr')))

    # mux a,v
    args = ['mux', '-f', '-ld', dst, ','.join([a_type, v_type])]
    args += a_splits
    args += v_splits
    parsed = marm.cli.arg_parser.parse_args(map(str, args))
    parsed.cmd(parsed)
    marm.FFProbe([dst.strpath])()

    # mux v,a
    args = ['mux', '-f', '-ld', dst, ','.join([v_type, a_type])]
    args.extend(sorted(tmpdir.visit('v-*.{0}'.format('mjr'))))
    args.extend(sorted(tmpdir.visit('a-*.{0}'.format('mjr'))))
    parsed = marm.cli.arg_parser.parse_args(map(str, args))
    parsed.cmd(parsed)
    marm.FFProbe([dst.strpath])()



import pytest

import marm


@pytest.mark.parametrize(
    'file_name,packet_type,expected', [
        ('sonic-a.mjr', marm.opus.OpusRTPPacket, 5996),
        ('sonic-v.mjr', marm.vp8.VP8RTPPacket, 9654),
    ]
)
def test_read_mjr(fixtures, file_name, packet_type, expected):
    path = fixtures.join(file_name)
    mjr = marm.mjr.MJRRTPPacketReader(path.open('rb'), packet_type=packet_type)
    assert sum(1 for _ in mjr) == expected



"""
Helpers for interacting w/ ffmpeg and friends.
"""
from datetime import timedelta
import json
import logging
import re
import subprocess


logger = logging.getLogger(__name__)


class Error(RuntimeError):

    def __init__(self, process, stdout, stderr):
        message = '"{0}" failed w/ return code {1}'.format(
            process.line, process.returncode
        )
        super(Error, self).__init__(message)
        self.line = process.line
        self.returncode = process.returncode
        self.stdout = stdout
        self.stderr = stderr


class Process(subprocess.Popen):

    #: Location of executable.
    bin = None

    #: Log level to use.
    log_level = logging.DEBUG

    #: Error type.
    Error = Error

    def __init__(self, args, **kwargs):
        args = [self.bin] + map(str, args)
        self.line = subprocess.list2cmdline(args)
        super(Process, self).__init__(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            **kwargs
        )

    def __call__(self, error='raise'):
        logger.debug('%s ...', self.line)
        stdout, stderr = self.communicate()
        log_level = (logging.ERROR if self.returncode != 0 else self.log_level)
        logging.log(
            log_level, '%s = %s - \n%s', self.line, self.returncode, stderr,
        )
        if self.returncode != 0:
            self._rejected(stdout, stderr)
            if error == 'raise':
                raise self.Error(self, stdout, stderr)
        else:
            self._fulfilled(stdout, stderr)
        return self.returncode

    def _fulfilled(self, stdout, stderr):
        pass

    def _rejected(self, stdout, stderr):
        pass


class FFMPEG(Process):
    """
    `subprocess.Popen` for executing `ffmpeg` and interpreting results.
    """

    # Process

    bin = 'ffmpeg'


class FFProbeResult(dict):
    """
    `dict` for parsing `FFProbe` results.
    """

    @classmethod
    def loads(cls, raw):

        def _for_mapping(obj):
            for k, v in obj.iteritems():
                if isinstance(v, dict):
                    _for_mapping(v)
                elif isinstance(v, (list, tuple)):
                    _for_sequence(v)
                elif isinstance(v, basestring):
                    obj[k] = cls._parse_string(v)
                

        def _for_sequence(obj):
            for k, v in enumerate(obj):
                if isinstance(v, dict):
                    _for_mapping(v)
                elif isinstance(v, (list, tuple)):
                    _for_sequence(v)
                elif isinstance(v, basestring):
                    obj[k] = cls._parse_string(v)

        obj = cls(json.loads(raw))
        _for_mapping(obj)
        return obj

    # internals
    
    @classmethod
    def _parse_string(cls, txt):
        for r, p in cls._VALUE_PARSERS:
            m = r.match(txt)
            if m:
                return p(m, txt)
        return txt

    _VALUE_PARSERS = [
        (re.compile(r'^\d+:\d+$'), lambda m, v: tuple(map(int, v.split(':')))),
        (re.compile(r'^\d+/\d+$'), lambda m, v: tuple(map(int, v.split('/')))),
        (re.compile(r'^\d+$'), lambda m, v: int(v)),
        (re.compile(r'^\d+\.\d+$'), lambda m, v: float(v)),
        (re.compile(r'^N\/A$'), lambda m, v: None),
    ]


class FFProbe(Process):
    """
    `subprocess.Popen` for executing `ffprobe` and interpreting results, e.g.:

    .. code:: python

        # run it
        ffprobe = FFProbe([
            '-show_format',
            '-show_streams',
            '-count_frames',
            '-count_packets',
            'mpegts_v1-1443977277-1443977282-00000.ts'
        ])
        ffprobe()

        # succeeded
        assert ffprobe.returncode == 0
        
        # and here are the results
        assert isinstance(ffprobe.result, FFProbe.Result)
        assert isinstance(ffprobe.result['format'], dict)
        assert isinstance(ffprobe.result['streams'], list)
        assert isinstance(ffprobe.result['stream'][0], dict)

    """

    Result = FFProbeResult

    #: Parsed stdout as `FFProbeResult`.
    result = None
    
    def __init__(self, *args, **kwars):
        args[0].extend([
            '-print_format', 'json',
        ])
        super(FFProbe, self).__init__(*args, **kwars)


    @classmethod
    def for_duration(cls, path):
        probe = cls(['-show_format', path])
        probe()
        return timedelta(seconds=probe.result['format']['duration'])

    # Process

    bin = 'ffprobe'

    def _fulfilled(self, stdout, stderr):
        self.result = FFProbeResult.loads(stdout)



from __future__ import division

import ctypes
import StringIO

from . import rtp


class VP8RTPPayload(rtp.RTPVideoPayloadMixin, rtp.RTPPayload):

    def __init__(self, *args, **kwargs):
        if args:
            if kwargs:
                raise TypeError('Mixed positional and keyword arguments')
            elif len(args) > 1:
                raise TypeError('Expected single \'buf\' arg')
            self.unpack(args[0])
        elif kwargs:
            if len(kwargs) == 1 and 'buf' in kwargs:
                self.unpack(kwargs['buf'])
            else:
                self.desc = VP8RTPPayloadDescriptor()
                self.desc_x = VP8RTPPayloadDescriptorX()
                self.desc_i = VP8RTPPayloadDescriptorI()
                self.desc_l = VP8RTPPayloadDescriptorL()
                self.desc_tk = VP8RTPPayloadDescriptorTK()
                self.data = ''
                for k, v in kwargs.iteritems():
                    if k not in ('desc', 'desc_x', 'desc_i', 'desc_l', 'desc_tk', 'data',):
                        raise TypeError('Unexpected keyword argument \'{0}\''.format(k))
                    setattr(self, k, v)

    @property
    def header(self):
        # NOTE: part of frame data so we don't unpack
        if not self.is_start_of_frame:
            raise ValueError('Not start of frame.')
        b = self.data[:ctypes.sizeof(VP8Header)]
        return VP8Header.from_buffer_copy(b)

    @property
    def key_header(self):
        # NOTE: part of frame data so we don't unpack
        if not self.is_start_of_frame:
            raise ValueError('Not start of frame.')
        if not self.header.is_key_frame:
            raise ValueError('Not key frame.')
        b = self.data[ctypes.sizeof(VP8Header):ctypes.sizeof(VP8Header) + ctypes.sizeof(VP8KeyFrameHeader)]
        return VP8KeyFrameHeader.from_buffer_copy(b)

    # rtp.RTPPayload

    def pack(self, fo=None):
        if fo is None:
            fo, value = StringIO.StringIO(), True
        else:
            value = False
        fo.write(buffer(self.desc))
        if self.desc.x:
            fo.write(buffer(self.desc_x))
        if self.desc_x and self.desc_x.i:
            fo.write(buffer(self.desc_i))
        if self.desc_x and self.desc_x.l:
            fo.write(buffer(self.desc_l))
        if self.desc_x and (self.desc_x.t or self.desc_x.k):
            fo.write(buffer(self.desc_tk))
        fo.write(self.data)
        if value:
            return fo.getvalue()

    def unpack(self, buf):
        # descriptor
        desc = VP8RTPPayloadDescriptor.from_buffer_copy(buf)
        buf = buf[ctypes.sizeof(desc):]

        # descriptor x
        if desc.x:
            desc_x = VP8RTPPayloadDescriptorX.from_buffer_copy(buf)
            buf = buf[ctypes.sizeof(desc_x):]
        else:
            desc_x = None

        # descriptor i
        if desc_x and desc_x.i:
            desc_i = VP8RTPPayloadDescriptorI.from_buffer_copy(buf)
            buf = buf[desc_i.size:]
        else:
            desc_i = None

        # descriptor l
        if desc_x and desc_x.l:
            desc_l = VP8RTPPayloadDescriptorL.from_buffer_copy(buf)
            buf = buf[ctypes.sizeof(desc_l):]
        else:
            desc_l = None

        # descriptor tk
        if desc_x and (desc_x.t or desc_x.k):
            desc_tk = VP8RTPPayloadDescriptorTK.from_buffer_copy(buf)
            buf = buf[ctypes.sizeof(desc_tk):]
        else:
            desc_tk = None

        # data
        data = buf

        self.desc = desc
        self.desc_x = desc_x
        self.desc_i = desc_i
        self.desc_l = desc_l
        self.desc_tk = desc_tk
        self.data = data

    # RTPVideoMixin

    @property
    def is_start_of_frame(self):
        return self.desc.s == 1 and self.desc.pid == 0

    @property
    def is_key_frame(self):
        return self.header.is_key_frame if self.is_start_of_frame else False

    @property
    def width(self):
        return self.key_header.width

    @property
    def height(self):
        return self.key_header.height


class VP8RTPPayloadDescriptor(ctypes.BigEndianStructure):
    """
    https://tools.ietf.org/html/draft-ietf-payload-vp8-16#section-4.2
    """

    _fields_ = [
        ('x', ctypes.c_uint8, 1),
        ('r', ctypes.c_uint8, 1),
        ('n', ctypes.c_uint8, 1),
        ('s', ctypes.c_uint8, 1),
        ('pid', ctypes.c_uint8, 4),
    ]


class VP8RTPPayloadDescriptorX(ctypes.BigEndianStructure):
    """
    https://tools.ietf.org/html/draft-ietf-payload-vp8-16#section-4.2
    """

    _fields_ = [
        ('i', ctypes.c_uint8, 1),
        ('l', ctypes.c_uint8, 1),
        ('t', ctypes.c_uint8, 1),
        ('k', ctypes.c_uint8, 1),
        ('rsv', ctypes.c_uint8, 4),
    ]


class VP8RTPPayloadDescriptorI(ctypes.BigEndianStructure):
    """
    https://tools.ietf.org/html/draft-ietf-payload-vp8-16#section-4.2
    """

    _fields_ = [
        ('m', ctypes.c_uint8, 1),
        ('pictureid0', ctypes.c_uint8, 7),
        ('pictureid1', ctypes.c_uint8, 8),
    ]

    @property
    def size(self):
        return ctypes.sizeof(self) - (0 if self.m else 1)


class VP8RTPPayloadDescriptorL(ctypes.BigEndianStructure):
    """
    https://tools.ietf.org/html/draft-ietf-payload-vp8-16#section-4.2
    """

    _fields_ = [
        ('tl0picidx', ctypes.c_uint8, 8),
    ]


class VP8RTPPayloadDescriptorTK(ctypes.BigEndianStructure):
    """
    https://tools.ietf.org/html/draft-ietf-payload-vp8-16#section-4.2
    """

    _fields_ = [
        ('tid', ctypes.c_uint8, 3),
        ('y', ctypes.c_uint8, 1),
        ('keyidx', ctypes.c_uint8, 4),
    ]


class VP8Header(ctypes.LittleEndianStructure):
    """
    https://tools.ietf.org/html/rfc6386#section-9.1
    """

    _fields_ = [
        ('p', ctypes.c_uint8, 1),
        ('ver', ctypes.c_uint8, 3),
        ('show', ctypes.c_uint8, 1),
        ('size0', ctypes.c_uint8, 3),
        ('size1', ctypes.c_uint8),
        ('size2', ctypes.c_uint8),
    ]

    @property
    def is_key_frame(self):
        return self.p == 0

    @property
    def size(self):
        return self.size0 + (self.size1 << 3) + (self.size2 << 11)


class VP8KeyFrameHeader(ctypes.LittleEndianStructure):
    """
    https://tools.ietf.org/html/rfc6386#section-9.1
    """

    _pack_ = 1

    _fields_ = [
        ('start_code0', ctypes.c_uint8),
        ('start_code1', ctypes.c_uint8),
        ('start_code2', ctypes.c_uint8),
        ('horz', ctypes.c_uint16),
        ('vert', ctypes.c_uint16),
    ]

    sync_code = (0x9d, 0x01, 0x2a)

    @property
    def is_synced(self):
        return (
            self.start_code0 == self.sync_code[0] and
            self.start_code1 == self.sync_code[1] and
            self.start_code2 == self.sync_code[2]
        )

    @property
    def width(self):
        return self.horz & 0x3fff

    @property
    def width_scaling(self):
        return self.horz >> 14

    @property
    def height(self):
        return self.vert & 0x3fff

    @property
    def height_scaling(self):
        return self.vert >> 14


class VP8RTPPacket(rtp.RTPTimeMixin, rtp.RTPPacket):
    """
    https://tools.ietf.org/html/draft-ietf-payload-vp8-16
    """

    # RTPPacket

    type = rtp.RTPPacket.VIDEO_TYPE

    payload_type = VP8RTPPayload

    # RTPTimeMixin

    clock_rate = 90000



import os
import struct

from . import rtp


class MJRRTPPacketReader(rtp.RTPPacketReader):
    """
    Iterates and indexes `RTPPacket`s from an MJR formatted file.
    """

    def __init__(self, *args, **kwargs):
        super(MJRRTPPacketReader, self).__init__(*args, **kwargs)
        self.type = read_header(self.fo)
        self.org = self.fo.tell()

    # rtp.RTPPacketReader
    
    def index(self, restore=True):
        org = pos = self.fo.tell()
        try:
            while True:
                skip_packet(self.fo)
                yield pos
                pos = self.fo.tell()
        except ValueError, ex:
            if not is_eof(ex):
                raise
        if restore:
            self.fo.seek(org)
    
    def __iter__(self):

        def pkts():
            for buf in read_packets(self.fo):
                pkt = self.packet_type(buf)
                if not self.packet_filter(pkt):
                    continue
                yield pkt

        self.fo.seek(self.org)
        return pkts()


rtp.RTPPacketReader.register('mjr', MJRRTPPacketReader)


MARKER = 'MEETECHO'

AUDIO_TYPE = 'audio'

VIDEO_TYPE = 'video'


def read_header(fo):
    read_marker(fo)
    type_ = read_string(fo)
    if type_ not in (AUDIO_TYPE, VIDEO_TYPE):
        raise ValueError('Unsupported type "{0}".'.format(type_))
    return type_


def read_string(fo, length=None):
    if length is None:
        b = fo.read(2)
        if len(b) != 2:
            raise ValueError('Failed to read string length at {0}.'.format(fo.tell()))
        l, = struct.unpack('>H', b)
        return read_string(fo, l)
    b = fo.read(length)
    if len(b) != length:
        raise ValueError('Failed to read {0} length string at {1}.'.format(length, fo.tell()))
    return b

def read_marker(fo):
    b = read_string(fo, len(MARKER))
    if b != MARKER:
        raise ValueError('Invalid marker "{0}" != "{1}"'.format(b, 'MEETCHO'))
    return b


def read_packet(fo):
    read_marker(fo)
    b = fo.read(2)
    if len(b) != 2:
        raise ValueError('Failed to read string length at {0}.'.format(fo.tell()))
    length, = struct.unpack('>H', b)
    buf = fo.read(length)
    if len(buf) != length:
        raise ValueError('Failed to read {0} length string at {1}.'.format(length, fo.tell()))
    return buf


def read_packets(fo):
    try:
        while True:
            yield read_packet(fo)
    except ValueError, ex:
        if not is_eof(ex):
            raise
        # eof


def skip_packet(fo):
    read_marker(fo)
    b = fo.read(2)
    if len(b) != 2:
        raise ValueError('Failed to read string length at {0}.'.format(fo.tell()))
    length, = struct.unpack('>H', b)
    pos = fo.tell()
    fo.seek(length, os.SEEK_CUR)
    if fo.tell() - pos != length:
        raise ValueError('Failed to seek past {0} length string at {1}.'.format(length, fo.tell()))
    return


def is_eof(ex):
    return (
        'Failed to read' in ex.message or
        'Failed to seek' not in ex.message
    )


def write_marker(fo):
    fo.write(MARKER)


def write_string(fo, buf):
    fo.write(struct.pack('>H', len(buf)))
    fo.write(buf)


def write_header(fo, type_):
    write_marker(fo)
    write_string(fo, type_)


def write_packet(fo, pkt):
    write_marker(fo)
    buf = pkt.pack()
    write_string(fo, buf)



import collections

import pytest

import marm


class frame_filter(marm.FrameFilter):

    def __init__(self):
        super(frame_filter, self).__init__()
        self.c = collections.Counter()


class keep_all(frame_filter):

    def __call__(self, frame):
        return self.KEEP_ALL


class keep_n(frame_filter):

    def __init__(self, n):
        super(keep_n, self).__init__()
        self.n = n

    def __call__(self, frame):
        self.c[frame.stream_index] += 1
        return (
            self.KEEP
            if self.c[frame.stream_index] <= self.n
            else self.DROP
        )


class drop_n(frame_filter):

    def __init__(self, n):
        super(drop_n, self).__init__()
        self.n = n

    def __call__(self, frame):
        self.c[frame.stream_index] += 1
        return (
            self.DROP
            if self.c[frame.stream_index] <= self.n
            else self.KEEP
        )


@pytest.mark.parametrize(
    'v_store,v_pkt,v_enc,a_store,a_pkt,a_enc,fmt,filter,counts', [
        ('sonic-v.mjr', marm.vp8.VP8RTPPacket, 'libvpx',
         'sonic-a.mjr', marm.opus.OpusRTPPacket, 'libopus',
         'mkv',
         keep_all(),
         {0: {'packets': 3555, 'frames': 3555},
          1: {'packets': 5996, 'frames': 5996}}),
        ('sonic-v.mjr', marm.vp8.VP8RTPPacket, 'libvpx',
         'sonic-a.mjr', marm.opus.OpusRTPPacket, 'libopus',
         'mkv',
         keep_n(10),
         {0: {'packets': 10, 'frames': 10},
          1: {'packets': 10, 'frames': 10}}),
        ('sonic-v.mjr', marm.vp8.VP8RTPPacket, 'libvpx',
         'sonic-a.mjr', marm.opus.OpusRTPPacket, 'libopus',
         'mkv',
         drop_n(10),
         {0: {'packets': 3545, 'frames': 3523},
          1: {'packets': 5986, 'frames': 5986}}),
    ]
)
def test_remux_filter(
        fixtures,
        tmpdir,
        v_store, v_pkt, v_enc,
        a_store, a_pkt, a_enc,
        fmt,
        filter,
        counts):
    # mux
    m1 = tmpdir.join('1.{0}'.format(fmt))
    pytest.mux(
        m1,
        fixtures.join(v_store), v_pkt, v_enc,
        fixtures.join(a_store), a_pkt, a_enc,
    )

    # remux it w/ filter
    m2 = tmpdir.join('2.{0}'.format(fmt))
    with m1.open('rb') as i, m2.open('wb') as o:
        marm.frame.remux(o, i, filter)

    # counts
    p = marm.FFProbe([
        '-count_packets',
        '-count_frames',
        '-show_streams',
        m2.strpath,
    ])
    p()
    assert counts == dict([(s['index'], {
        'packets': s['nb_read_packets'],
        'frames': s['nb_read_frames'],
    }) for s in p.result['streams']])



from __future__ import division

import abc
import collections
import logging
import os
import StringIO
import struct

from . import ext


logger = logging.getLogger(__name__)


class Frame(object):
    """
    Represents a frame as a tuple of:
    
    - pts
    - flags
    - data
    
    corresponding to `libavcodec.AVPacket`.
    """

    FLAG_KEY_FRAME = 1 << 0  # AV_PKT_FLAG_KEY
    FLAG_CORRUPT = 1 << 1  # AV_PKT_FLAG_CORRUPT

    def __init__(self, *args, **kwargs):
        if args:
            if kwargs:
                raise TypeError('Mixed positional and keyword arguments')
            elif len(args) > 1:
                raise TypeError('Expected single \'buf\' arg')
            self.unpack(args[0])
        elif kwargs:
            if len(kwargs) == 1 and 'buf' in kwargs:
                self.unpack(kwargs['buf'])
            else:
                self.pts, self.flags, self.data = 0, 0, ''
                for k, v in kwargs.iteritems():
                    if k not in ('pts', 'flags', 'data'):
                        raise TypeError('Unexpected keyword argument \'{0}\''.format(k))
                    setattr(self, k, v)

    @property
    def is_key_frame(self):
        return self.flags & self.FLAG_KEY_FRAME != 0

    @property
    def is_corrupt(self):
        return self.flags & self.FLAG_CORRUPT != 0

    def unpack(self, buf):
        fo = StringIO.StringIO(buf) if isinstance(buf, basestring) else buf
        fmt = '=qii'
        size = struct.calcsize(fmt)
        b = fo.read(size)
        if len(b) != size:
            raise ValueError('Failed to read {0} bytes for "{1}" at {2}.'.format(size, fmt, fo.tell()))
        pts, flags, data_size = struct.unpack(fmt, b)
        data = fo.read(data_size)
        if len(data) != data_size:
            raise ValueError('Failed to read {0} byte packet data at {1}.'.format(data_size, fo.tell()))
        self.pts, self.flags, self.data = pts, flags, data

    def pack(self, fo=None):
        fo, value = StringIO.StringIO(), True if fo is None else fo, False
        fo.write(struct.pack('=qqii', self.pts, self.flags, len(self.data)))
        fo.write(self.data)
        if value:
            return fo.getvalue()


class Frames(collections.Iterator):
    """
    Depacketizes `Frame`s assuming each packet is a frame.
    """

    def __init__(self, packets, pts_offset=0, pts='msecs', flags=0, frame_type=Frame):
        self.packets = iter(packets)
        self.pts = pts
        self.pts_offset = pts_offset
        self.flags = flags
        self.frame_type = frame_type

    # collections.Iterator

    def __iter__(self):
        return self

    def next(self):
        packet = self.packets.next()
        pts = int(getattr(packet, self.pts) + self.pts_offset)
        return self.frame_type(
            pts=pts,
            flags=self.flags,
            data=packet.data.data
        )

class VideoFrame(Frame):
    """
    Convenience specialization of `Frame` for video.
    """

    PIX_FMT_NONE = -1  # AV_PIX_FMT_NONE
    PIX_FMT_YUV420P = 0  # AV_PIX_FMT_YUV420P
    PIX_FMT_YUYV422 = 1  # AV_PIX_FMT_YUYV422
    PIX_FMT_RGB24 = 2  # AV_PIX_FMT_RGB24


class AudioFrame(Frame):
    """
    Convenience specialization of `Frame` for audio.
    """

    CHANNEL_LAYOUT_MONO = ext.AV_CH_LAYOUT_MONO
    CHANNEL_LAYOUT_STEREO = ext.AV_CH_LAYOUT_STEREO

    def __init__(self, *args, **kwargs):
        kwargs['flags'] = kwargs.pop('flags', 0) | Frame.FLAG_KEY_FRAME
        super(AudioFrame, self).__init__(*args, **kwargs)


class VideoFrames(collections.Iterator):
    """
    Depacketizes `VideoFrame`s. There may be more than one packet/frame and to
    group them into a single `VideoFrame` the packet data should support:
    
    - `RTPVideoPayloadMixin`.
    
    """

    def __init__(self, packets, pts_offset=0, pts='msecs'):
        self.packets = iter(packets)
        self.pts = pts
        try:
            self.packet = self.packets.next()
        except StopIteration:
            self.packet = None
        if self.packet:
            self.packet, self.start_frame_offset = self._seek_start_frame()
        if self.packet:
            self.packet, self.key_frame_offset = self._seek_key_frame()
        self.pts_offset = pts_offset

    def _seek_start_frame(self):
        packet = self.packet

        # already there
        if packet.data.is_start_of_frame:
            return packet, 0

        # forward
        offset = 0
        for packet in self.packets:
            if packet.data.is_start_of_frame:
                return packet, offset
            offset += 1
            logger.debug('dropping non-frame-start packet')

        # nothing
        return None, offset

    def _seek_key_frame(self):
        packet = self.packet

        # already there
        if packet.data.is_key_frame:
            return packet, 0

        # forward
        offset = 0
        for packet in self.packets:
            if packet.data.is_start_of_frame and packet.data.is_key_frame:
                return packet, offset
            if not packet.data.is_start_of_frame:
                offset += 1
                logger.debug('dropping non-key-frame-start packet')

        return None, offset

    def _read_frame(self):
        first = self.packet

        # meta
        pts, flags = int(getattr(first, self.pts) + self.pts_offset), 0
        if first.data.header.is_key_frame:
            flags |= VideoFrame.FLAG_KEY_FRAME

        # data
        fo = StringIO.StringIO()
        fo.write(first.data.data)
        count = 1
        for self.packet in self.packets:
            if self.packet.data.is_start_of_frame:
                break
            fo.write(self.packet.data.data)
            count += 1
        else:
            self.packet = None
        data = fo.getvalue()

        return VideoFrame(pts=pts, flags=flags, data=data)

    # collections.Iterator

    def __iter__(self):
        return self

    def next(self):
        if self.packet is None:
            raise StopIteration()
        if not self.packet.data.is_start_of_frame:
            logger.debug('dropping non-frame-start packet')
            for self.packet in self.packet:
                if self.packet.data.is_start_of_frame:
                    break
                logger.debug('dropping non-frame-start packet')
            else:
                raise StopIteration()
        packet = self._read_frame()
        if packet is None:
            raise StopIteration()
        return packet


# libav* codec.
Codec = ext.Codec


# Iterates `ext.Codec`s registered with libav*.
codecs = ext.codecs


def gen_audio(fo=None, encoder_name=None, **kwargs):
    """
    Generates encoded audio frames which is useful for testing.
    """
    encoder_name = encoder_name or format_ext(fo, 'fo')[1:]
    if fo is None:
        fo = StringIO.StringIO()
        ext.generate_audio(
            fo,
            encoder_name,
            header=False,
            **kwargs)
        fo.seek(0)
        return read_frames(fo)
    return ext.generate_audio(fo, encoder_name, **kwargs)


def gen_video(fo, encoder_name=None, **kwargs):
    """
    Generates encoded video frames which is useful for testing.
    """
    encoder_name = encoder_name or format_ext(fo, 'fo')[1:]
    return ext.generate_video(fo, encoder_name, **kwargs)


# libav* format.
Format = ext.Format


# Iterates Formats registered with libav*.
formats = ext.output_formats


def mux(fo,
        video_profile=None,
        video_packets=None,
        audio_profile=None,
        audio_packets=None,
        format_extension=None):
    """
    Muxes encoded video and audio frames (i.e. codec packets) into a container.
    """
    format_extension = format_extension or format_ext(fo, 'fo')
    ext.mux(
        fo,
        format_extension,
        v_profile=video_profile,
        v_packets=video_packets,
        a_profile=audio_profile,
        a_packets=audio_packets,
    )


# libav* `libavcodec.AVPacket` proxy.
FrameProxy = ext.Packet


class FrameFilter(object):
    
    KEEP = ext.FILTER_KEEP
    DROP = ext.FILTER_DROP
    KEEP_ALL = ext.FILTER_KEEP_ALL
    DROP_ALL = ext.FILTER_DROP_ALL

    @abc.abstractmethod
    def __call__(self, frame_proxy):
        pass

    @classmethod
    def range(cls, *args, **kwargs):
        return FrameRange(*args, **kwargs)


class FrameRange(FrameFilter):

    def __init__(self, stream_index, b, e=None, count=0):
        self.stream_index = stream_index
        self.b, self.e, self.count = b, e, count

    def __call__(self, frame):
        if frame.stream_index != self.stream_index:
            return self.KEEP
        idx = self.count
        self.count += 1
        if self.b is not None and idx < self.b:
            return self.DROP
        if self.e is not None and self.e < idx:
            return self.DROP
        return self.KEEP

    def shift(self, v):
        return type(self)(
            self.stream_index,
            self.b + v if self.b is not None else self.b,
            self.e + v if self.e is not None else self.e,
        )


def remux(
        out_fo,
        in_fo,
        filter=None,
        out_format_extension=None,
        in_format_extension=None,
        **kwargs):
    """
    Re-muxes encoded video and audio frames (i.e. codec packets) into another
    container.
    """
    out_format_extension = out_format_extension or format_ext(out_fo, 'out_fo', )
    in_format_extension = in_format_extension or format_ext(in_fo, 'in_fo',)
    return ext.remux(
        out_fo,
        out_format_extension,
        in_fo,
        in_format_extension,
        filter,
        **kwargs
    )


def last_mpegts_ccs(in_fo, in_format_extension=None):
    in_format_extension = in_format_extension or format_ext(in_fo, 'in_fo')
    return ext.last_mpegts_ccs(in_fo, in_format_extension)


def segment(
        out_template,
        out_format_name,
        in_fo,
        in_format_extension=None,
        **kwargs):
    """
    Splits encoded video and audio into smaller segments.
    """
    in_format_extension = in_format_extension or format_ext(in_fo, 'in_fo')
    return ext.segment(
        out_template,
        out_format_name,
        in_fo,
        in_format_extension,
        **kwargs
    )


def format_ext(fo, tag):
    _, ext = os.path.splitext(fo.name)
    if not ext:
        raise ValueError('{0}.name "{1}" has no extension.'.format(tag, fo.name))
    return ext


class VideoHeader(collections.namedtuple('VideoHeader', [
        'encoder_name',
        'pix_fmt',
        'width',
        'height',
        'bit_rate',
        'frame_rate',
    ])):

    def pack(self):
        fmt = '=B{0}sB{1}siiiii'.format(len('video'), len(self.encoder_name))
        buf = struct.pack(fmt,
            len('video'), b'video',
            len(self.encoder_name), self.encoder_name.encode('ascii'),
            self.pix_fmt,
            self.width,
            self.height,
            self.bit_rate,
            self.frame_rate,
        )
        return buf

    @classmethod
    def unpack(cls, fo):
        return cls(*((read_string(fo),) + read_struct(fo, 'iiiii')))


class AudioHeader(collections.namedtuple('AudioHeader', [
        'encoder_name',
        'bit_rate',
        'sample_rate',
        'channel_layout',
    ])):

    def pack(self):
        fmt = '=B{0}sB{1}siiQ'.format(len('audio'), len(self.encoder_name))
        buf = struct.pack(fmt,
            len('audio'), 'audio',
            len(self.encoder_name), self.encoder_name,
            self.bit_rate,
            self.sample_rate,
            self.channel_layout,
        )
        return buf

    @classmethod
    def unpack(cls, fo):
        return cls(*((read_string(fo),) + read_struct(fo, 'iiQ')))


def read_header(fo):
    t = read_string(fo)
    if t == 'video':
        h = VideoHeader.unpack(fo)
    elif t == 'audio':
        h = AudioHeader.unpack(fo)
    else:
        raise ValueError('Unsupported type "{0}".'.format(t))
    return h


def read_string(fo, length=None):
    if length is None:
        b = fo.read(1)
        if len(b) != 1:
            raise ValueError('Failed to read string length at {0}.'.format(fo.tell()))
        l, = struct.unpack('=B', b)
        return read_string(fo, l)
    b = fo.read(length)
    if len(b) != length:
        raise ValueError('Failed to read {0} length string at {1}.'.format(length, fo.tell()))
    return b


def read_struct(fo, fmt):
    size = struct.calcsize(fmt)
    b = fo.read(size)
    if len(b) != size:
        raise ValueError('Failed to read {0} bytes for "{1}" at {2}.'.format(size, fmt, fo.tell()))
    return struct.unpack(fmt, b)


def read_frames(fo, frame_type=Frame):
    try:
        while True:
            frame = frame_type(fo)
            yield frame
    except ValueError, ex:
        if not is_eof(ex):
                raise
            # eof


def is_eof(ex):
    return (
        'Failed to read' in ex.message or
        'Failed to seek' not in ex.message
    )


def write_string(fo, buf):
    fo.write(struct.pack('=B{0}s'.format(len(buf)), len(buf), buf))


def write_header(fo, header):
    if isinstance(header, VideoHeader):
        write_string(fo, 'video')
        fo.write(header.pack())
    elif isinstance(header, AudioHeader):
        write_string(fo, 'audio')
        fo.write(header.pack())
    else:
        raise TypeError(
            'Invalid header type {0}'.format(type(header).__name__)
        )


def monotonic(frames):
    pts = None
    delta = 1
    for frame in frames:
        if pts is not None:
            if frame.pts <= pts:
                frame.pts = pts + delta
            else:
                delta = frame.pts - pts
        pts = frame.pts
        yield frame



import StringIO

import mock
import pytest

import marm


@pytest.mark.parametrize(
    'encoder_name,duration,width,height,frame_rate,frame_count', [
        ('mpeg4', 11, 320, 240, 25, 275),
        ('mpeg4', 5, 640, 480, 30, 150),
    ])
def test_gen_video_frames(
            tmpdir,
            encoder_name,
            duration,
            width,
            height,
            frame_rate,
            frame_count,
        ):
    path = tmpdir.join('v.{0}'.format(encoder_name))
    with path.open('wb') as fo:
        marm.frame.gen_video(
                fo,
                duration=duration,
                width=width,
                height=height,
                frame_rate=frame_rate
            )
    with path.open('rb') as fo:
        header = marm.frame.read_header(fo)
        assert header == marm.frame.VideoHeader(
            encoder_name=encoder_name,
            pix_fmt=0,
            width=width,
            height=height,
            bit_rate=400000,
            frame_rate=frame_rate,
        )
        c = 0
        for _ in marm.frame.read_frames(fo):
            c += 1
        assert c == frame_count


def test_gen_video_frames_write_error(tmpdir):

    class MyException(Exception):

        pass

    encoder, duration, width, height, frame_rate, = 'mpeg4', 5, 640, 480, 30
    io = StringIO.StringIO()
    with mock.patch.object(io, 'write') as p:
        p.side_effect = MyException('bam')
        with pytest.raises(MyException) as ei:
            marm.frame.gen_video(
                io,
                duration=duration,
                encoder_name=encoder,
                pix_fmt=0,
                width=width,
                height=height,
                bit_rate=400000,
                frame_rate=frame_rate,
            )
        assert 'bam' in ei.value


@pytest.mark.parametrize(
    'encoder_name,duration,bit_rate,sample_rate,channel_layout,frame_count', [
        ('flac', 11, 96000, 48000, marm.AudioFrame.CHANNEL_LAYOUT_STEREO, 115),
        ('flac', 5, 128000, 44100, marm.AudioFrame.CHANNEL_LAYOUT_MONO, 48),
    ])
def test_gen_audio_frames(
        tmpdir,
        encoder_name,
        duration,
        bit_rate,
        sample_rate,
        channel_layout,
        frame_count):
    path = tmpdir.join('a.{0}'.format(encoder_name))
    with path.open('wb') as fo:
        marm.frame.gen_audio(
            fo,
            duration=duration,
            bit_rate=bit_rate,
            sample_rate=sample_rate,
            channel_layout=channel_layout
        )

    with path.open('rb') as fo:
        header = marm.frame.read_header(fo)
        assert header == marm.frame.AudioHeader(
            encoder_name=encoder_name,
            bit_rate=bit_rate,
            sample_rate=sample_rate,
            channel_layout=channel_layout,
        )
        c = sum(1 for _ in marm.frame.read_frames(fo))
        assert abs(c - frame_count) < 2


def test_gen_audio_frames_write_error(tmpdir):

    class MyException(Exception):

        pass

    encoder, duration, bit_rate, sample_rate = 'libopus', 10, 96000, 48000
    io = StringIO.StringIO()
    with mock.patch.object(io, 'write') as p:
        p.side_effect = MyException('boom')
        with pytest.raises(MyException) as ei:
            marm.frame.gen_audio(
                io,
                encoder_name=encoder,
                duration=duration,
                bit_rate=bit_rate,
                sample_rate=sample_rate,
            )
        assert 'boom' in ei.value



import re
import setuptools.command.test
import subprocess


class PyTest(setuptools.command.test.test):

    user_options = []

    def finalize_options(self):
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest

        pytest.main(self.test_args)

version = (
    re
    .compile(r".*__version__ = '(.*?)'", re.S)
    .match(open('marm/__init__.py').read())
    .group(1)
)

packages = setuptools.find_packages('.', exclude=('test',))

scripts = [
    'script/marm'
]

ext_modules = [
    setuptools.Extension(
        'marm.ext',
        ['marm/ext/gen.c',
         'marm/ext/mux.c',
         'marm/ext/remux.c',
         'marm/ext/scan.c',
         'marm/ext/segment.c',
         'marm/ext/util.c',
         'marm/ext/mpegts.c',
         'marm/ext/ext.pyx'],
        include_dirs=['marm/ext/'],
        extra_compile_args=(
            ['-g', '-O0'] +
            subprocess.check_output(['pkg-config', '--cflags', 'libavformat']).strip().split() +
            subprocess.check_output(['pkg-config', '--cflags', 'libavcodec']).strip().split() +
            subprocess.check_output(['pkg-config', '--cflags', 'libswscale']).strip().split() +
            subprocess.check_output(['pkg-config', '--cflags', 'libavutil']).strip().split()
        ),
        extra_link_args=(
            ['-Wl,-Bsymbolic'] +
            subprocess.check_output(['pkg-config', '--libs', 'libavformat']).strip().split() +
            subprocess.check_output(['pkg-config', '--libs', 'libavcodec']).strip().split() +
            subprocess.check_output(['pkg-config', '--libs', 'libswscale']).strip().split() +
            subprocess.check_output(['pkg-config', '--libs', 'libavutil']).strip().split()
        )
    )
]
try:
    import Cython.Build
    ext_modules = Cython.Build.cythonize(ext_modules)
except ImportError:
    pass


extras_require = {
    'test': [
        'mock >=1.3,<2',
        'pytest >=2.5.2,<3',
        'pytest-cache >=1.0,<2',
        'pytest-cov >=1.7,<2',
        'pytest-pep8 >=1.0.6,<2',
    ],
}

setuptools.setup(
    name='marm',
    version=version,
    url='https://github.com/mayfieldrobotics/marm/',
    author='Mayfield Robotics',
    author_email='dev+marm@mayfieldrobotics.com',
    license='MIT',
    description='Muxing archived media (and more!).',
    long_description=open('README.rst').read(),
    packages=packages,
    scripts=scripts,
    ext_modules=ext_modules,
    platforms='any',
    install_requires=[
        'dpkt >=1.8,<2',
    ],
    tests_require=extras_require['test'],
    extras_require=extras_require,
    cmdclass={'test': PyTest},
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



import marm


def test_codecs():
    assert isinstance(marm.frame.codecs().next(), marm.frame.Codec)
    assert any(c.name == 'mpeg4' for c in marm.frame.codecs())
    assert any(c.name == 'flac' for c in marm.frame.codecs())



import subprocess

import pytest

import marm


@pytest.mark.parametrize(
    ('duration,'
     'v_encoder,v_width,v_height,v_frame_rate,'
     'a_encoder,a_bit_rate,a_sample_rate,a_channel_layout,'
     'fmt,fmt_name'), [
        (5,
         'mpeg4', 320, 240, 25,
         'flac', 96000, 48000, marm.frame.AudioFrame.CHANNEL_LAYOUT_STEREO,
         'mkv', 'matroska'),
    ]
)
def test_mux_gen(
        tmpdir,
        duration,
        v_encoder, v_width, v_height, v_frame_rate,
        a_encoder, a_bit_rate, a_sample_rate, a_channel_layout,
        fmt, fmt_name):
    v_path = tmpdir.join('v.{0}'.format(v_encoder))
    a_path = tmpdir.join('a.{0}'.format(a_encoder))
    m_path = tmpdir.join('m.{0}'.format(fmt))

    with v_path.open('wb') as fo:
        marm.frame.gen_video(
            fo,
            duration=duration,
            width=v_width,
            height=v_height,
            frame_rate=v_frame_rate
        )

    with a_path.open('wb') as fo:
        marm.frame.gen_audio(
            fo,
            duration=duration,
            bit_rate=a_bit_rate,
            sample_rate=a_sample_rate,
            channel_layout=a_channel_layout,
        )

    with m_path.open('wb') as fo, \
            v_path.open('rb') as v_fo, \
            a_path.open('rb') as a_fo:
        # a
        a_hdr = marm.frame.read_header(a_fo)
        a_frames = marm.frame.read_frames(a_fo)

        # v
        v_hdr = marm.frame.read_header(v_fo)
        v_frames = marm.frame.read_frames(v_fo)

        # mux them
        marm.frame.mux(
            fo,
            audio_profile={
                'encoder_name': a_hdr.encoder_name,
                'bit_rate': a_hdr.bit_rate,
                'sample_rate': a_hdr.sample_rate,
                'channel_layout': a_hdr.channel_layout,
            },
            audio_packets=a_frames,
            video_profile={
                'encoder_name': v_hdr.encoder_name,
                'pix_fmt': v_hdr.pix_fmt,
                'width': v_hdr.width,
                'height': v_hdr.height,
                'frame_rate': v_hdr.frame_rate,
                'bit_rate': v_hdr.bit_rate,
            },
            video_packets=v_frames,
        )

    probe = marm.FFProbe([
        '-show_format',
        '-show_streams',
        m_path.strpath,
    ])
    probe()
    assert len(probe.result['streams']) == 2
    assert fmt_name in probe.result['format']['format_name'].split(',')


def check_output(cmd):
    try:
        return subprocess.check_output(cmd, shell=True), None
    except Exception, ex:
        return None, ex


@pytest.mark.parametrize(
    ('v_store,v_enc,a_store,a_enc,dur,fmt'), [
        ('sonic-v.mjr', 'libvpx', 'sonic-a.mjr', 'libopus', 10.0, 'mkv'),
    ]
)
def test_concat_muxed(
        tmpdir,
        fixtures,
        pool,
        ffmpeg,
        v_store, v_enc,
        a_store, a_enc,
        dur,
        fmt):
    # split v
    v_path = fixtures.join(v_store)
    v_parts = 0
    v_pkts = marm.mjr.MJRRTPPacketReader(
        str(v_path),
        packet_type=marm.vp8.VP8RTPPacket
    )
    for i, part in enumerate(marm.rtp.split_packets(v_pkts, duration=dur)):
        t_path = tmpdir.join('v-{0}.mjr'.format(i + 1))
        with t_path.open('wb') as t_fo:
            marm.mjr.write_header(t_fo, 'video')
            for pkt in part:
                marm.mjr.write_packet(t_fo, pkt)
        v_parts += 1

    # split a
    a_path = fixtures.join(a_store)
    a_parts = 0
    a_pkts = marm.mjr.MJRRTPPacketReader(
        str(a_path),
        packet_type=marm.opus.OpusRTPPacket
    )
    for i, part in enumerate(marm.rtp.split_packets(a_pkts, duration=dur)):
        t_path = tmpdir.join('a-{0}.mjr'.format(i + 1))
        with t_path.open('wb') as t_fo:
            marm.mjr.write_header(t_fo, 'audio')
            for pkt in part:
                marm.mjr.write_packet(t_fo, pkt)
        a_parts += 1

    # cursor for split parts
    v_cur = marm.rtp.RTPCursor([
            str(tmpdir.join('v-{0}.mjr'.format(i + 1)))
            for i in range(v_parts)
        ],
        marm.mjr.MJRRTPPacketReader,
        packet_type=marm.vp8.VP8RTPPacket,
    )
    a_cur = marm.rtp.RTPCursor([
            str(tmpdir.join('a-{0}.mjr'.format(i + 1)))
            for i in range(a_parts)
        ],
        marm.mjr.MJRRTPPacketReader,
        packet_type=marm.opus.OpusRTPPacket,
    )

    # probe
    v_prof = pytest.probe(v_cur)
    a_prof = pytest.probe(a_cur)

    # calculate mux points
    mux_points = []
    assert v_parts == a_parts
    for i in range(v_parts):
        # begin
        v_cur.seek((i, 0))
        v_b_drop = 0
        a_cur.seek((i, 0))
        a_b_drop = 0
        a_b_delay = 0
        a_b = a_cur.tell()
        v_pkt = v_cur.current()
        if v_cur.prev_key_frame():
            v_cur.prev()
            v_b = v_cur.tell()
            a_b_delay = int((v_pkt.msecs - v_cur.current().msecs))
            v_b_drop = (
                v_cur.count((i, 0), lambda pkt: pkt.data.is_start_of_frame) - 1
            )
        elif v_pkt.data.is_key_frame:
            v_b = v_cur.tell()
        elif v_cur.next_key_frame():
            v_cur.prev()
            v_b = v_cur.tell()
            a_cur.fastforward(v_pkt.msecs - v_cur.current().msecs)
            a_b = a_cur.tell()
        else:
            assert not (
                'Could not find key-frame in "{0}" part "{1}"'
                .format(v_path, i + 1)
            )

        # end
        v_cur.seek((i, -1))
        v_pkt = v_cur.current()
        if not v_pkt.data.is_start_of_frame:
            if v_cur.prev_start_of_frame() is None:
                assert not (
                    'Could not find prev start-of-frame in "{0}" part "{1}"'
                    .format(v_path, i + 1)
                )
            v_cur.prev()
        v_e = v_cur.tell()
        a_e = (i + 1, 0)

        mux_points.append((v_b, v_b_drop, v_e, a_b, a_b_drop, a_b_delay, a_e))

    # mux parts
    for i, (v_b, _, v_e, a_b, _, a_b_delay, a_e) in enumerate(mux_points):
        # packets
        v_cur.seek(v_b)
        v_pkts = v_cur.slice(v_e)
        a_cur.seek(a_b)
        a_pkts = a_cur.slice(a_e)

        # mux
        p_path = tmpdir.join('mux-{0}.{1}'.format(i + 1, fmt))
        with p_path.open('wb') as fo:
            marm.frame.mux(
                fo,
                video_profile={
                    'encoder_name': v_enc,
                    'pix_fmt': marm.VideoFrame.PIX_FMT_YUV420P,
                    'width': v_prof['width'],
                    'height': v_prof['height'],
                    'frame_rate': v_prof['frame_rate'],
                    'bit_rate': v_prof['bit_rate'],
                    'time_base': (1, 1000),
                },
                video_packets=marm.VideoFrames(v_pkts),
                audio_profile={
                    'encoder_name': a_enc,
                    'bit_rate': a_prof['bit_rate'],
                    'sample_rate': a_prof['sample_rate'],
                    'channel_layout': a_prof['channel_layout'],
                    'time_base': (1, 1000),
                },
                audio_packets=marm.Frames(a_pkts, a_b_delay),
            )

    # xcode parts
    xcodes = []
    for i, point in enumerate(mux_points):
        v_b, v_b_drop, v_e, a_b, a_b_drop, _, a_e = point
        p_path = tmpdir.join('mux-{0}.{1}'.format(i + 1, fmt))
        x_path = tmpdir.join('xcode-{0}.{1}'.format(i + 1, fmt))
        cmd = [
            ffmpeg,
            '-i', p_path,
            '-codec:v', v_enc,
            '-filter:v',
            '"select=gte(n\\, {0}), setpts=PTS-STARTPTS"'.format(v_b_drop),
            '-codec:a', a_enc,
            '-filter:a',
            '"aselect=gte(n\\, {0}), asetpts=PTS-STARTPTS"'.format(a_b_drop),
            x_path,
        ]
        xcodes.append(' '.join(map(str, cmd)))
    for _, ex in pool.imap(check_output, xcodes):
        assert ex is None

    # concat xcoded parts
    c_txt_path = tmpdir.join('concat.txt')
    c_txt_path.write('\n'.join([
        'file {0}'.format(tmpdir.join('xcode-{0}.{1}'.format(i + 1, fmt)))
        for i in range(len(mux_points))
    ]))
    c_path = tmpdir.join('concat.{0}'.format(fmt))
    cmd = [
        ffmpeg,
        '-f', 'concat',
        '-i', str(c_txt_path),
        '-c', 'copy',
        str(c_path),
    ]
    subprocess.check_call(args=' '.join(cmd), shell=True)


@pytest.mark.parametrize(
    ('v_store,v_pt,v_enc,a_store,a_pt,a_enc,fmt,count'), [
        ('sonic-v.mjr', marm.vp8.VP8RTPPacket, 'libvpx',
         'sonic-a.mjr', marm.opus.OpusRTPPacket, 'libopus',
         'mkv',
         100),
    ]
)
def test_mux_next_packet_error(
        tmpdir,
        fixtures,
        v_store, v_pt, v_enc,
        a_store, a_pt, a_enc,
        fmt,
        count):
    # v
    v_src = fixtures.join(v_store)
    v_pkts = marm.rtp.RTPCursor([v_src.strpath], packet_type=v_pt)
    v_prof = pytest.probe(v_pkts)

    # a
    a_src = fixtures.join(a_store)
    a_pkts = marm.rtp.RTPCursor([a_src.strpath], packet_type=a_pt)
    a_prof = pytest.probe(a_pkts)

    # mux

    class MyException(Exception):

        pass

    local = {'call_count': 0}

    def v_pkts_wrapper():
        for i, v_pkt in enumerate(v_pkts):
            local['call_count'] += 1
            if i > count:
                raise MyException('poop')
            yield v_pkt

    m_path = tmpdir.join('mux.{0}'.format(fmt))
    with pytest.raises(MyException) as ei:
        with m_path.open('wb') as fo:
            marm.frame.mux(
                fo,
                video_profile={
                    'encoder_name': v_enc,
                    'pix_fmt': marm.VideoFrame.PIX_FMT_YUV420P,
                    'width': v_prof['width'],
                    'height': v_prof['height'],
                    'frame_rate': v_prof['frame_rate'],
                    'bit_rate': v_prof['bit_rate'],
                    'time_base': (1, 1000),
                },
                video_packets=marm.VideoFrames(v_pkts_wrapper()),
                audio_profile={
                    'encoder_name': a_enc,
                    'bit_rate': a_prof['bit_rate'],
                    'sample_rate': a_prof['sample_rate'],
                    'channel_layout': a_prof['channel_layout'],
                    'time_base': (1, 1000),
                },
                audio_packets=marm.Frames(a_pkts)
            )
    assert 'poop' in ei.value
    assert local == {'call_count': count + 2}
